(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Extra operations on nellists\<close>

text \<open>
The operations ndropns, nkfilter, nleast and nidx are defined for nellists together with a library
of lemmas. 
\<close>

theory NELListKFilter 
 imports   NELList
  

begin

subsection \<open>ndropns\<close>

primcorec ndropns :: " 'a nellist \<Rightarrow> 'a nellist nellist "
 where "ndropns nell = 
        (case nell of (NNil b) \<Rightarrow> (NNil (NNil b)) |
          (NCons x nell') \<Rightarrow> (NCons (NCons x nell') (ndropns nell'))) "

simps_of_case ndropns_code [code, simp, nitpick_simp]: ndropns.code

lemma ndropns_simps [simp]:
 shows nhd_ndropns: " \<not> is_NNil nell \<Longrightarrow> nhd (ndropns nell) = nell "
  and  ndropns_NCons: " ntl (ndropns nell) = (case nell of (NNil b) \<Rightarrow> (NNil (NNil b)) |
                                                (NCons x nell') \<Rightarrow> ndropns nell')  "
by (auto simp add: nellist.case_eq_if)   
   (metis ndropns_code(1) nellist.collapse(1) nellist.sel(4))

lemma ndropns_nnth:
 assumes " i \<le> nlength nell" 
  shows " nnth (ndropns nell) i = ndropn i nell"
using assms 
proof (induction i arbitrary: nell)
case 0
then show ?case 
 proof (cases nell)
 case (NNil x1)
 then show ?thesis by (simp add: nnth_NNil)
 next
 case (NCons x21 x22)
 then show ?thesis by simp
 qed
next
case (Suc i)
then show ?case 
 proof (cases nell)
 case (NNil x1)
 then show ?thesis by (simp add: nnth_NNil)
 next
 case (NCons x21 x22)
 then show ?thesis using Suc by (simp add: Suc_ile_eq)
 qed
qed

lemma ndropns_nlength:
 " nlength (ndropns nell) = (nlength nell)"
by (coinduction arbitrary: nell rule: enat_coinduct)
   (case_tac nell, auto)

lemma in_nset_ndropns:
 " nell \<in> nset(ndropns nellx) \<longleftrightarrow> (\<exists> i. i \<le> nlength nellx \<and> nell = ndropn i nellx)"
by (metis in_nset_conv_nnth ndropns_nlength ndropns_nnth)

lemma nset_ndropns:
 " nset (ndropns nell) = { ndropn i nell | i. i \<le> nlength nell}"
using in_nset_conv_nnth[of _ "ndropns nell"] nset_conv_nnth[of "ndropns nell"]
using in_nset_ndropns[of  _ nell] by auto 

lemma nmap_first_ndropns:
 " nmap (\<lambda> nell. nnth nell 0) (ndropns nell) = nell"
by (coinduction arbitrary: nell)
   (case_tac nell, auto simp add: nnth_NNil)

lemma ndropn_ndropns:
 assumes " i \<le> nlength(ndropns nell)"
 shows " ndropn i (ndropns nell) = ndropns (ndropn i nell)"
using assms
proof (coinduction arbitrary: nell i)
case (Eq_nellist ia nellx)
then show ?case 
   proof -
    have 1: "enat nellx \<le> nlength (ndropns ia) \<Longrightarrow>
             is_NNil (ndropn nellx (ndropns ia)) = is_NNil (ndropns (ndropn nellx ia))"
      by (simp add: is_NNil_ndropn ndropns_nlength) 
    have 2: "enat nellx \<le> nlength (ndropns ia) \<Longrightarrow>
             (is_NNil (ndropn nellx (ndropns ia)) \<longrightarrow>
              is_NNil (ndropns (ndropn nellx ia)) \<longrightarrow> 
              nlast (ndropn nellx (ndropns ia)) = nlast (ndropns (ndropn nellx ia)))"
      by (metis dual_order.antisym ndropn_eq_NNil ndropns.disc(2) ndropns.simps(3) ndropns_nlength
          ndropns_nnth nellist.case_eq_if nellist.collapse(1) the_enat.simps) 
    have 3: "enat nellx \<le> nlength (ndropns ia) \<Longrightarrow>
             (\<not> is_NNil (ndropn nellx (ndropns ia)) \<longrightarrow>
              \<not> is_NNil (ndropns (ndropn nellx ia)) \<longrightarrow>
              nhd (ndropn nellx (ndropns ia)) = nhd (ndropns (ndropn nellx ia)))"
      by (metis Nat.add_0_right ndropn_nnth ndropns.disc(1) ndropns_nlength ndropns_nnth nhd_conv_nnth
          nhd_ndropns) 
    have 4: "enat nellx \<le> nlength (ndropns ia) \<Longrightarrow>
             (\<not> is_NNil (ndropn nellx (ndropns ia)) \<longrightarrow>
              \<not> is_NNil (ndropns (ndropn nellx ia)) \<longrightarrow>
              (\<exists>nell i.
                ntl (ndropn nellx (ndropns ia)) = ndropn i (ndropns nell) \<and>
                ntl (ndropns (ndropn nellx ia)) = ndropns (ndropn i nell) \<and> 
                enat i \<le> nlength (ndropns nell))) "
      by (metis Suc_ile_eq is_NNil_ndropn ndropn_Suc_conv_ndropn ndropns_code(2) ndropns_nlength
          nellist.sel(5) order.order_iff_strict) 
    show ?thesis 
     using "1" "2" "3" "4" Eq_nellist by blast
  qed
qed

lemma ndropns_nfilter_nnth:
 assumes " i \<le> nlength (nfilter P (ndropns nell))"
         " \<exists> nelly \<in> nset(ndropns nell). P nelly"
 shows   " P (nnth (nfilter P (ndropns nell)) i) "
using assms using nset_nfilter[of "ndropns nell" P] 
by (metis (full_types) Int_iff in_nset_conv_nnth mem_Collect_eq ) 

lemma nnth_zero_ndropn:
 " nnth (ndropn n nell) 0 = nnth nell n"
by simp

lemma in_nset_ndropns_nhd:
 "x \<in> nset nell \<longleftrightarrow>( \<exists> ys. x=(nnth ys 0) \<and> ys \<in> nset(ndropns nell))"
by auto
   (metis in_nset_conv_nnth ndropns_nlength nmap_first_ndropns nnth_nmap,
    metis Nat.add_0_right in_nset_conv_nnth in_nset_ndropns ndropn_nnth)

lemma nset_ndropns_nhd:
" nset nell = {(nnth ys 0) | ys. ys \<in> nset(ndropns nell) }"
by auto
   (meson in_nset_ndropns_nhd,
    metis in_nset_conv_nnth in_nset_ndropns nnth_zero_ndropn)

lemma nellist_all2_ndropnsI:
 " nellist_all2 A xs ys   \<Longrightarrow> nellist_all2 (nellist_all2 A) (ndropns xs) (ndropns ys)"
by (coinduction arbitrary: xs ys)
   (auto simp add: nellist.case_eq_if dest: nellist_all2_nhdD nellist_all2_is_NNilD 
    intro: nellist_all2_ntlI)

context includes lifting_syntax
begin
lemma ndropns_transfer2 [transfer_rule]:
 " (nellist_all2 A ===> nellist_all2 (nellist_all2 A)) ndropns ndropns"
unfolding rel_fun_def
by (auto intro: nellist_all2_ndropnsI  )

end

context
includes nellist.lifting
begin

lift_definition nkfilter :: " ('a \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow>'a nellist \<Rightarrow> nat nellist"
is "\<lambda> P n ll.   (if lnull(kfilter P n ll) then (iterates Suc n) else kfilter P n ll)"
by simp

lift_definition nleast :: " ('a \<Rightarrow> bool) \<Rightarrow> 'a nellist \<Rightarrow> nat "
is "\<lambda> P ll. lleast P ll"
by auto

lift_definition nidx :: " nat nellist \<Rightarrow> bool "
 is "\<lambda> ll. lidx ll"
by auto


subsection \<open>nkfilter\<close>

lemma nkfilter_NNil [simp]:
  shows "P b \<Longrightarrow> nkfilter P n (NNil b)  = (NNil n)"
by transfer auto 

lemma nkfilter_NCons [simp]:
 assumes  " (\<exists> x \<in> nset nell. P x)" 
 shows    " nkfilter P n (NCons x nell) = 
           (if P x then NCons n (nkfilter P (Suc n) nell) else nkfilter P (Suc n) nell)"
using assms
by transfer
   (auto simp add: kfilter_lnull_conv)

lemma nkfilter_NCons_a [simp]:
assumes " \<not>(\<exists> x \<in> nset nell. P x)"  
        "P x"
shows   " nkfilter P n (NCons x nell) = (NNil n)"
using assms
by transfer
   (auto simp add: kfilter_lnull_conv)

lemma nkfilter_nlength:
 assumes "\<exists> x \<in> nset nell. P x"
 shows   " nlength(nkfilter P n nell) = nlength(nfilter P nell)"
using assms
by transfer
   (auto simp add: kfilter_lnull_conv kfilter_llength)

lemma nkfilter_upperbound:
 assumes "\<exists> x \<in> nset nell. P x"
         " i \<le> nlength (nkfilter P n nell)"
 shows   " nnth (nkfilter P n nell) i \<le> n + nlength nell" 
using assms
by transfer
   (auto, 
    simp add: kfilter_lnull_conv,
    metis co.enat.exhaust_sel iadd_Suc_right iless_Suc_eq kfilter_upperbound llength_eq_0)

lemma nkfilter_lowerbound:
 assumes "\<exists> x \<in> nset nell. P x"
         " i \<le> nlength (nkfilter P n nell)"
 shows   " n \<le> nnth (nkfilter P n nell) i "
 using assms
by transfer
  (auto,
   metis co.enat.collapse iless_Suc_eq kfilter_lnth_n_zero_a llength_eq_0)

lemma nkfilter_mono:
 assumes "\<exists> x \<in> nset nell. P x"
         " (Suc i) \<le> nlength (nkfilter P n nell)"
 shows   " nnth (nkfilter P n nell) i < nnth (nkfilter P n nell) (Suc i)"
using assms
by transfer
   (auto,
    metis diff_Suc_1 eSuc_epred epred_enat epred_le_epredI lidx_kfilter_gr iless_Suc_eq leD lessI 
    llength_eq_0 min.cobounded1 min_def the_enat.simps)

lemma nkfilter_nfilter:
 assumes "\<exists> x \<in> nset nell. P x"
         " i \<le> nlength (nkfilter P n nell)"
 shows   " (nnth nell ((nnth (nkfilter P n nell) i) - n)) = (nnth (nfilter P nell) i)"
using assms
apply transfer
proof (auto simp add: kfilter_lnull_conv split:if_split_asm)
 fix nella :: "'a llist" and Pa :: "'a \<Rightarrow> bool" and ia :: nat and na :: nat and x :: 'a
 assume a1: "x \<in> lset nella"
 assume a2: "Pa x"
 assume a3: "enat ia \<le> epred (llength (kfilter Pa na nella))"
 assume a4: "\<not> lnull nella"
 have f5: "\<forall>e ea. \<not> (e::enat) \<le> ea \<or> min e ea = e"
 by (simp add: min_def_raw)
 have f6: "\<forall>n p na as. \<not> enat n < llength (kfilter p na (as::'a llist)) \<or>
             enat (lnth (kfilter p na as) n) < enat na + llength as"
 by (meson kfilter_upperbound)
 have f7: "min (enat ia) (epred (llength (kfilter Pa na nella))) = enat ia"
 using f5 a3 by blast
 then have f8: "(enat ia < llength (kfilter Pa 0 nella)) = 
               (enat (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) < 
                llength (kfilter Pa na nella))"
 by (simp add: kfilter_llength)
 have f9: "\<forall>n p na as. \<not> enat n < llength (kfilter p na (as::'a llist)) \<or> 
              lnth (kfilter p na as) n - na = lnth (kfilter p 0 as) n"
 using kfilter_lnth_n_zero by blast
 have "eSuc (epred (llength nella)) = enat 0 + llength nella"
 using a4 by (metis co.enat.exhaust_sel gen_llength_def llength_code llength_eq_0)
 then have "enat (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) < 
           llength (kfilter Pa na nella) \<longrightarrow> 
           enat (lnth (kfilter Pa na nella) 
                      (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) - na)
           < eSuc (epred (llength nella))"
 using f9 f8 f7 f6 the_enat.simps by presburger
 then have f10: "enat (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) <
                 llength (kfilter Pa na nella) \<longrightarrow> 
                enat (lnth (kfilter Pa na nella) 
                           (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) - na)
                 \<le> epred (llength nella)"
 using iless_Suc_eq by blast
 have f11: "\<forall>n p na as. \<not> enat n < llength (kfilter p na as) \<or> lnth as (lnth (kfilter p na as) n - na) =
                (lnth (lfilter p as) n::'a)"
 using lfilter_kfilter by blast
 have f12: "the_enat (min (enat ia) (epred (llength (lfilter Pa nella)))) = 
           the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))"
 by (simp add: kfilter_llength)
 have "\<not> lnull (kfilter Pa na nella)"
 using a1 a2 kfilter_not_lnull_conv by auto
 then have "enat (the_enat (min (enat ia) (epred (llength (kfilter Pa na nella))))) < 
           llength (kfilter Pa na nella)"
 using f7 a3 by (metis co.enat.exhaust_sel iless_Suc_eq llength_eq_0 the_enat.simps)
 then show "\<not> lnull nella \<Longrightarrow>
       enat ia \<le> epred (llength (kfilter Pa na nella)) \<Longrightarrow>
       x \<in> lset nella \<Longrightarrow>
       Pa x \<Longrightarrow>
       lnth nella (the_enat (min (enat (lnth (kfilter Pa na nella) ia - na)) (epred (llength nella)))) =
       lnth (lfilter Pa nella) (the_enat (min (enat ia) (epred (llength (lfilter Pa nella)))))"
 using f12 f11 f10 f5 by simp
qed

lemma in_nkfilter_nset:
 assumes "\<exists> x \<in> nset nell. P x"
 shows " x \<in> nset(nkfilter P n nell) \<longleftrightarrow> x \<in> { n+i | i. i \<le> nlength nell \<and> P (nnth nell i)} "
using assms
by transfer
   (auto simp add: kfilter_lnull_conv  min_def,
    metis co.enat.exhaust_sel gen_llength_def iless_Suc_eq kfilter_holds kfilter_llength_n_zero 
       kfilter_lnth_n_zero kfilter_lowerbound kfilter_upperbound ldistinct_Ex1 ldistinct_kfilter 
       le_add_diff_inverse llength_code llength_eq_0,
    metis add.commute co.enat.exhaust_sel iless_Suc_eq kfilter_holds_not_a llength_eq_0)

lemma nkfilter_nset:
 assumes "\<exists> x \<in> nset nell. P x"
 shows " nset(nkfilter P n nell) = { n+i | i. i \<le> nlength nell \<and> P (nnth nell i)} "
using assms in_nkfilter_nset[of nell P _ n] by blast  

lemma nlength_nkfilter_le [simp]:
 assumes "\<exists> x \<in> nset nell. P x"
shows "nlength (nkfilter P n nell) \<le> nlength nell"
using assms
by transfer
    (auto simp add: kfilter_lnull_conv epred_le_epredI kfilter_llength llength_lfilter_ile)

lemma nkfilter_nleast:
 assumes "\<exists> x \<in> nset xs. P x"
 shows   " nnth (nkfilter P n xs) 0 = n+(nleast P xs)"
using assms
by transfer
   (auto simp add: kfilter_not_lnull_conv  kfilter_lnull_conv,
    metis enat_min_eq_0_iff kfilter_lnth_zero kfilter_lnull_conv the_enat_0 zero_enat_def)

lemma ndistinct_nkfilter:
 assumes "\<exists> x \<in> nset nell. P x"
 shows   "ndistinct(nkfilter P n nell)"
using assms
by transfer
  (auto simp add: kfilter_lnull_conv ldistinct_kfilter)

lemma nkfilter_ndropn_nset:
 assumes "\<exists> x \<in> nset (ndropn k nell). P x"
         " k \<le> nlength nell " 
 shows " nset(nkfilter P n (ndropn k nell)) = { n+i|i. i\<le>nlength nell - k \<and> P (nnth nell (k+i))}"
using assms nkfilter_nset[of "(ndropn k nell)" P n] 
ndropn_nnth[of _ nell k] by auto 

lemma nkfilter_ndropn_nset_b:
 assumes "\<exists> x \<in> nset (ndropn k nell). P x"
         " k \<le> nlength nell " 
 shows " nset(nkfilter P n (ndropn k nell)) = { n+ i |i. i\<le> nlength nell - k  \<and> P (nnth nell (i+k))}"
proof -
 have 1: "{ n+i|i. i\<le>nlength nell - k \<and> P (nnth nell (k+i))} = 
          { n+ i |i. i\<le> nlength nell - k  \<and> P (nnth nell (i+k))}" 
   using assms by (auto simp add: add.commute)
 show ?thesis using assms by (simp add: "1"  nkfilter_ndropn_nset)
qed

lemma nfilter_nkfilter_ntaken_nlength_0:
 assumes " P (nnth nell ( (nnth (nkfilter P n nell) k) -n )) "
         " k \<le> nlength (nfilter P nell) "
 shows   " (\<exists> x \<in> nset nell. P x) "
using assms using exists_nnth_nset by blast

lemma nkfilter_nlength_n_zero:
 assumes " (\<exists> x \<in> nset nell. P x) "
 shows   " nlength(nkfilter P n nell) = nlength(nkfilter P 0 nell)"
using assms by (simp add: nkfilter_nlength)

lemma nkfilter_nnth_n_zero:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " k \<le> nlength(nkfilter P n nell) "
 shows   " (nnth (nkfilter P n nell) k) - n = (nnth (nkfilter P 0 nell) k) "
using assms 
by transfer
  (auto split: if_split_asm,
   metis kfilter_lnull_conv,
   metis kfilter_lnull_conv,
   metis co.enat.exhaust_sel iless_Suc_eq kfilter_llength_n_zero kfilter_lnth_n_zero llength_eq_0 
   min.orderE the_enat.simps)

lemma nkfilter_n_zero:
 assumes " (\<exists> x \<in> nset nell. P x) "
 shows " (nkfilter P n nell) = nmap (\<lambda>i. i+n) (nkfilter P 0 nell)"
proof -
 have 1: "nlength(nkfilter P n nell) = nlength (nmap (\<lambda>i. i+n) (nkfilter P 0 nell))" 
   using assms nkfilter_nlength_n_zero by fastforce
 have 2: "\<And> k. k \<le> nlength (nkfilter P n nell) \<longrightarrow> 
           nnth (nkfilter P n nell) k = nnth (nmap (\<lambda>i. i+n) (nkfilter P 0 nell)) k"
   using 1 assms nkfilter_nnth_n_zero[of nell P _ n]  
   by (metis diff_add nkfilter_lowerbound nlength_nmap nnth_nmap)
 show ?thesis by (simp add: "1" "2" nellist_eq_nnth_eq)
qed

lemma nkfilter_n_zero_a:
 assumes " (\<exists> x \<in> nset nell. P x) "
 shows " (nkfilter P 0 nell) = nmap (\<lambda>i. i-n) (nkfilter P n nell)"
proof -
 have 1: "nlength(nkfilter P 0 nell) = nlength (nmap (\<lambda>i. i-n) (nkfilter P n nell))" 
   by (metis assms nkfilter_nlength_n_zero nlength_nmap)
 have 2: "\<And>k. k\<le> nlength(nkfilter P 0 nell) \<longrightarrow> 
               nnth (nkfilter P 0 nell) k = nnth (nmap (\<lambda>i. i-n) (nkfilter P n nell)) k"
   by (simp add: "1" assms nkfilter_nnth_n_zero)
 show ?thesis by (simp add: "1" "2" nellist_eq_nnth_eq)
qed

lemma nkfilter_holds:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " y \<in> nset(nkfilter P n nell)"
 shows   " P (nnth nell (y-n))"
using assms  in_nkfilter_nset[of nell P]  by force

lemma nkfilter_holds_not:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " y \<in> {i+n |i. i\<le> nlength nell} - (nset (nkfilter P n nell))"
 shows  " \<not> P (nnth nell (y-n)) "
using assms nkfilter_nset[of nell P n] by auto 

lemma nkfilter_holds_a:
 assumes " (\<exists> x \<in> nset nell. P x) "
         "i \<le> nlength nell"
         " (i+n) \<in> nset(nkfilter P n nell)"
 shows " P (nnth nell i) "
using assms nkfilter_holds[of nell P "i+n" n ] by simp

lemma nkfilter_holds_not_a:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " i \<le> nlength nell"
         " P (nnth nell i) "
 shows   " (i+n) \<in>  nset(nkfilter P n nell)"
using assms by (simp add: in_nkfilter_nset)

lemma nkfilter_holds_b:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " i \<le> nlength nell"
 shows   " (i+n) \<in> nset(nkfilter P n nell) = P (nnth nell i)"
using assms by (meson nkfilter_holds_a nkfilter_holds_not_a)

lemma nkfilter_holds_c:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " n \<le> i" 
         " i-n \<le> nlength nell"
 shows   " i \<in> nset(nkfilter P n nell) = P (nnth nell (i-n))"
using assms 
by (metis diff_add idiff_enat_enat nkfilter_holds_b) 

lemma nkfilter_holds_not_b:
  assumes " (\<exists> x \<in> nset nell. P x) "
          " n \<le> i"
          " i-n \<le> nlength nell"
  shows   " i\<notin> nset (nkfilter P n nell) = (\<not> P (nnth nell (i-n))) "
using assms 
by (simp add: nkfilter_holds_c)

lemma nkfilter_disjoint_nset_coset:
 assumes " (\<exists> x \<in> nset nell. P x) "
 shows " ({i+n|i. i \<le> nlength nell} - nset(nkfilter P n nell)) \<inter> nset (nkfilter P n nell) = {}"
using assms by (simp add: inf.commute)

lemma nidx_less:
 assumes " nidx nell"
         " Suc(n+k) \<le> nlength nell"
 shows   " nnth nell n < nnth nell (Suc(n+k))"
using assms
apply transfer
proof auto 
 fix nella :: "nat llist" and na :: nat and ka :: nat
 assume a1: "lidx nella"
 assume a2: "enat (Suc (na + ka)) \<le> epred (llength nella)"
 assume a3: "\<not> lnull nella"
 have f4: "min na (na + ka) = na"
 by simp
 have f5: "\<forall>n na. min (enat n) (enat na) = enat (min n na)"
 using min_enat_simps(1) by blast
 have f6: "\<forall>e ea. if (e::enat) \<le> ea then min e ea = e else min e ea = ea"
 by (simp add: min_def_raw)
 have "enat (na + ka) \<le> epred (llength nella)"
 using a2 Suc_ile_eq by force
 then have "min (enat na) (epred (llength nella)) = enat na"
 by (metis f4 f6 min.bounded_iff min_enat_simps(1))
 then show " \<not> lnull nella \<Longrightarrow>
       lidx nella \<Longrightarrow>
       enat (Suc (na + ka)) \<le> epred (llength nella) \<Longrightarrow>
       lnth nella (the_enat (min (enat na) (epred (llength nella)))) < lnth nella (Suc (na + ka))"
 using f6 a3 a2 a1
 by (metis (no_types) co.enat.exhaust_sel iless_Suc_eq lidx_less llength_eq_0 the_enat.simps)
qed

lemma nidx_less_eq:
 assumes " nidx nell"
         " k \<le> j"
         " j \<le> nlength nell"
 shows   " nnth nell k \<le> nnth nell j"
using assms
by transfer
 (auto, metis co.enat.exhaust_sel dual_order.trans enat_ord_simps(1) iless_Suc_eq lidx_less_eq 
    llength_eq_0 min.orderE the_enat.simps)

lemma nidx_less_last:
 assumes " nidx nell"
         " Suc i < k"
         " nlength nell = (enat k) "
 shows   " nnth nell i < nnth nell (k-1)"
using assms less_imp_Suc_add nidx_less by fastforce

lemma nidx_gr_first:
 assumes " nidx nell"
         " 0<i"
         " i \<le> nlength nell"
 shows   " (nnth nell 0) < nnth nell i"
using assms nidx_less[of nell 0 "i-1"] 
by transfer auto

lemma nidx_expand:
 "nidx nell \<longleftrightarrow> (\<forall>i. (Suc i) \<le> nlength nell \<longrightarrow> nnth nell i < nnth nell (Suc i))"
by transfer
   (auto simp add: min_def Suc_ile_eq,
    metis Extended_Nat.eSuc_mono co.enat.exhaust_sel eSuc_enat lidx_def llength_eq_0,
    metis Suc_ile_eq co.enat.exhaust_sel iless_Suc_eq less_imp_le lidx_def llength_eq_0)

lemma nidx_nkfilter_expand:
 assumes " (\<exists> x \<in> nset nell. P x) "
         "(Suc i) \<le> nlength(nkfilter P n nell)"
 shows   " nnth (nkfilter P n nell) i < nnth (nkfilter P n nell) (Suc i)"
using assms by (simp add: nkfilter_mono)

lemma nidx_nkfilter:
 assumes " (\<exists> x \<in> nset nell. P x) "
 shows   "nidx (nkfilter P n nell)"
using assms nidx_nkfilter_expand[of nell P _ n]  nidx_expand[of "(nkfilter P n nell)" ] 
by blast

lemma nidx_nkfilter_gr_eq:
 assumes " (\<exists> x \<in> nset nell. P x) "
         " k \<le> j"
         " j \<le> nlength (nkfilter P n nell) " 
 shows   " nnth (nkfilter P n nell) k \<le> nnth (nkfilter P n nell) j"
using assms  nidx_nkfilter[of nell P n]  nidx_less_eq[of "nkfilter P n nell" k j]
by blast

lemma nidx_nkfilter_gr:
 assumes " (\<exists> x \<in> nset nell. P x) "    
 shows   "(\<forall> j. k<j \<and> j \<le> nlength (nkfilter P n nell) \<longrightarrow>
                 nnth (nkfilter P n nell) k < nnth(nkfilter P n nell) j) "
using assms nidx_nkfilter[of nell P n] nidx_less[of "nkfilter P n nell" ]
using less_imp_Suc_add by blast 

lemma nidx_nkfilter_less_eq:
 assumes " (\<exists> x \<in> nset nell. P x) "  
         "k \<le> nlength (nkfilter P n nell)"
 shows   "\<forall>j. j \<le> k \<longrightarrow> nnth (nkfilter P n nell) j \<le> nnth (nkfilter P n nell) k"
using assms by (simp add: nidx_nkfilter_gr_eq) 

lemma nkfilter_not_before:
  assumes " (\<exists> x \<in> nset nell. P x) " 
          " i < (nnth (nkfilter P 0 nell) 0) "
   shows  " \<not> P (nnth nell i) "
using assms
by transfer  
   (auto simp add: min_def split: if_split_asm,
    meson kfilter_not_before llength_eq_0 not_gr_zero,
    metis dual_order.strict_trans kfilter_not_before less_enatE llength_eq_0 not_gr_zero 
    not_le_imp_less the_enat.simps,
    meson le_less_linear less_enatE less_nat_zero_code,
    meson le_less_linear less_enatE not_less0)

lemma nkfilter_n_not_before:
  assumes " (\<exists> x \<in> nset (ndropn n nell). P x) "
          " n \<le> nlength nell"
          " n \<le> i"  
          " i < (nnth (nkfilter P n (ndropn n nell) ) 0) "
   shows  " \<not> P (nnth nell i) "
using assms
apply transfer
unfolding min_def 
using zero_enat_def
proof (auto split: if_split_asm)
 show "\<And>n nell P i x.
       enat 0 = 0 \<Longrightarrow>
       \<not> lnull nell \<Longrightarrow>
       enat n \<le> epred (llength nell) \<Longrightarrow>
       n \<le> i \<Longrightarrow>
       \<not> lnull (kfilter P n (ldropn n nell)) \<Longrightarrow>
       i < lnth (kfilter P n (ldropn n nell)) 0 \<Longrightarrow>
       x \<in> lset (ldropn n nell) \<Longrightarrow> P x \<Longrightarrow> enat i \<le> epred (llength nell) \<Longrightarrow> P (lnth nell i) \<Longrightarrow> False"
 by (metis co.enat.exhaust_sel iless_Suc_eq kfilter_n_not_before llength_eq_0 not_gr_zero) 
next
 fix na :: nat and nella :: "'b llist" and Pa :: "'b \<Rightarrow> bool" and ia :: nat and x :: 'b
 assume a1: "ia < lnth (kfilter Pa na (ldropn na nella)) 0"
 assume a2: "\<not> lnull (kfilter Pa na (ldropn na nella))"
 assume a3: "enat na \<le> epred (llength nella)"
 assume a4: "\<not> lnull nella"
 assume a5: "\<not> enat ia \<le> epred (llength nella)"
 assume a6: "Pa (lnth nella (the_enat (epred (llength nella))))"
 have f7: "\<forall>p n bs. lnull (kfilter p n (bs::'b llist)) \<or> lnth (kfilter p n bs) 0 = n + lleast p bs"
 by (meson kfilter_lnth_zero)
 then have f8: "lnth (kfilter Pa na (ldropn na nella)) 0 = na + lleast Pa (ldropn na nella)"
 using a2 by blast
 have f9: "0 < llength (kfilter Pa na (ldropn na nella))"
 using a2 by simp
 have f10: "na \<le> the_enat (epred (llength nella))" 
  by (metis a3 a5 enat_ord_code(4) enat_ord_simps(1) enat_the_enat less_imp_le)
 have f11: "the_enat (epred (llength nella)) < lnth (kfilter Pa na (ldropn na nella)) 0"
   by (metis a1 a5 dual_order.strict_trans enat_ord_simps(2) enat_ord_simps(3) enat_the_enat 
       not_le_imp_less)
 then show False using kfilter_n_not_before[of Pa na nella "(the_enat (epred (llength nella)))"]
   using f10 f11 
   by (metis a3 a4 a6 co.enat.exhaust_sel f9 iless_Suc_eq llength_eq_0)
qed

lemma nkfilter_not_after:
assumes " (\<exists> x \<in> nset nell. P x) " 
        " nnth (nkfilter P 0 nell) k < i "
        " nlength(nkfilter P 0 nell) = (enat k)"
        " i \<le> nlength nell"
 shows  " \<not> P (nnth nell i)"
using assms
by transfer
   (auto simp add: min_def split: if_split_asm,
    metis co.enat.exhaust_sel diff_Suc_1 eSuc_enat iless_Suc_eq kfilter_not_after llength_eq_0 
    not_gr_zero)

lemma nkfilter_n_not_after:
assumes " (\<exists> x \<in> nset (ndropn n nell). P x) "
        " n \<le> nlength nell"  
        " nnth (nkfilter P n (ndropn n nell)) k < i "
        " nlength(nkfilter P n (ndropn n nell)) = (enat k)"
        " i \<le> nlength nell"
shows  " \<not> P (nnth nell i)"
using assms 
by transfer
   (auto split: if_split_asm,
    metis diff_Suc_1 eSuc_enat eSuc_epred iless_Suc_eq kfilter_n_not_after llength_eq_0  not_gr_zero )

lemma nkfilter_not_between:
 assumes " (\<exists> x \<in> nset nell. P x) " 
         " nnth (nkfilter P 0 nell) k < i "
         " i < nnth (nkfilter P 0 nell) (Suc k)"
         " (Suc k) \<le> nlength (nkfilter P 0 nell)"
 shows   " \<not> P (nnth nell i) "
using assms
by transfer
   (auto simp add: min_def split: if_split_asm,
    metis co.enat.exhaust_sel iless_Suc_eq kfilter_not_between llength_eq_0,
    metis co.enat.exhaust_sel gen_llength_def iless_Suc_eq kfilter_upperbound le_less_trans 
    llength_code llength_eq_0 min.cobounded2 min.strict_order_iff min_enat_simps(1),
    meson Suc_ile_eq less_imp_le,
    meson Suc_ile_eq dual_order.strict_implies_order)

lemma ndistinct_Ex1:
 assumes "ndistinct nell"
         " x \<in> nset nell"
 shows   " \<exists>!i. i \<le> nlength nell \<and> (nnth nell i) = x"
using assms 
by transfer
   (auto,
    metis co.enat.exhaust_sel iless_Suc_eq ldistinct_Ex1 llength_eq_0 min.orderE the_enat.simps,
    metis co.enat.exhaust_sel iless_Suc_eq ldistinct_conv_lnth llength_eq_0)

lemma nidx_nset_eq:
 assumes " nidx nellx"
         " nidx nelly"
         "nset nellx = nset nelly"
 shows   "nellx = nelly"
using assms
by transfer
   (simp add: bi_unique_cr_nellist_help lidx_lset_eq nidx.rep_eq)

lemma ntaken_nset:
 assumes "k\<le> nlength nell"
 shows   " nset (ntaken k nell) = { (nnth nell i) | i. i \<le> k}"
using assms 
by (auto simp add: in_nset_conv_nnth)
   (meson ntaken_nnth, metis ntaken_nnth)

lemma ndropn_nset:
 assumes "k\<le> nlength nell"
 shows   " nset (ndropn k nell) = { (nnth nell i) | i. k \<le> i \<and> i \<le> nlength nell}"
using assms 
by (auto simp add: in_nset_conv_nnth)
   (metis enat_ile enat_ord_simps(1) le_add1 le_cases nnth_beyond,
    metis enat_ile le_cases le_iff_add ndropn_nlength ndropn_nnth nnth_beyond)

lemma nfilter_nkfilter_ntaken_nidx_a:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (ntaken k (nkfilter P n nell)) "
using assms 
by (simp add: nidx_expand nidx_nkfilter_gr ntaken_nnth)

lemma nfilter_nkfilter_ndropn_nidx_a:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (ndropn k (nkfilter P n nell)) "
using assms
by transfer
   (auto simp add: kfilter_lnull_conv,
    metis (no_types, lifting) gr_implies_not_zero kfilter_llength leI lfilter_kfilter_ldropn_lidx_a 
    lidx_def llength_eq_0 lnull_ldropn)

lemma nfilter_nkfilter_ntaken_nidx_b:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)))" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell)) "
 using assms nidx_nkfilter[of "(ntaken (nnth (nkfilter P 0 nell) k) nell)" P 0]
 by (metis nfinite_ntaken nset_nlast ntaken_nlast) 
 
lemma nfilter_nkfilter_ntaken_nidx_b_1:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell)) "
 using assms nfilter_nkfilter_ntaken_nidx_b[of P nell k]  nkfilter_holds[of nell P]  
by (metis in_nset_conv_nnth nkfilter_nlength nkfilter_nnth_n_zero)

lemma nfilter_nkfilter_ntaken_nidx_b_2:
 assumes "P (nnth nell ( (nnth (nkfilter P n nell) k) -n))" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P n (ntaken (nnth (nkfilter P n nell) k) nell)) "
 using assms nidx_nkfilter[of "(ntaken (nnth (nkfilter P n nell) k) nell)" P n]
 by (metis diff_le_self exists_nnth_nset ntaken_nnth) 
 
lemma nfilter_nkfilter_ntaken_nidx_b_3:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P n (ntaken (nnth (nkfilter P n nell) k) nell)) "
using assms nfilter_nkfilter_ntaken_nidx_b_2[of P nell n k] nkfilter_holds
by (metis in_nset_conv_nnth nkfilter_nlength)

lemma nfilter_nkfilter_ndropn_nidx_b:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)))" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k) nell)) "
using assms nidx_nkfilter[of "(ndropn (nnth (nkfilter P 0 nell) k) nell)" P 0]
by (metis diff_add diff_zero in_nset_conv_nnth ndropn_nnth zero_enat_def zero_le) 

lemma nfilter_nkfilter_ndropn_nidx_b_1:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k) nell)) "
 using assms nfilter_nkfilter_ndropn_nidx_b[of P nell k]  nkfilter_holds[of nell P] 
by (metis in_nset_conv_nnth nkfilter_nlength nkfilter_nnth_n_zero)

lemma nfilter_nkfilter_ndropn_nidx_b_2:
 assumes "P (nnth nell ( (nnth (nkfilter P n nell) k) -n))" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P ((nnth (nkfilter P n nell) k)-n) 
                            (ndropn ((nnth (nkfilter P n nell) k)-n) nell) ) "
proof -
 have 1: "enat (nnth (nkfilter P n nell) k) -n \<le> nlength nell"
   using nkfilter_upperbound[of nell P]  nkfilter_nnth_n_zero[of nell P k n]  assms
   by (metis gen_nlength_def idiff_enat_enat nfilter_nkfilter_ntaken_nlength_0 
       nkfilter_nlength nlength_code) 
 have 2: "nset (ndropn ((nnth (nkfilter P n nell) k)-n) nell) =
            { (nnth nell i) |i. ((nnth (nkfilter P n nell) k)-n) \<le> i \<and> i \<le> nlength nell } " 
   using ndropn_nset[of "(nnth (nkfilter P n nell) k)-n" nell]  1 by simp
 have 3: "\<exists>x\<in>nset (ndropn ((nnth (nkfilter P n nell) k)-n) nell). P x" 
   using "1" "2" assms(1) by auto
 show ?thesis using 3 
   nidx_nkfilter[of "(ndropn ((nnth (nkfilter P n nell) k)-n) nell)" P "(nnth (nkfilter P n nell) k)-n"]
   by blast
qed

lemma nfilter_nkfilter_ndropn_nidx_b_3:
 assumes "\<exists> x \<in> nset nell. P x" 
         " k \<le> nlength (nfilter P nell)" 
 shows   " nidx (nkfilter P ((nnth (nkfilter P n nell) k)-n) 
                            (ndropn ((nnth (nkfilter P n nell) k)-n) nell)) "
using assms nfilter_nkfilter_ndropn_nidx_b_2 
by (metis in_nset_conv_nnth nkfilter_holds nkfilter_nlength)

lemma ntaken_nkfilter_ntaken_nset_eq:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)) )" 
         "k \<le> nlength (nfilter P nell) "
 shows   " nset(ntaken k (nkfilter P 0 nell)) =
           nset(nkfilter P 0 (ntaken ((nnth (nkfilter P 0 nell) k)) nell)) "
proof -
 have 1: "( (nnth (nkfilter P 0 nell) k)) \<le> nlength nell "
   using assms nkfilter_upperbound[of nell P k 0]
   by (metis diff_zero gen_nlength_def nfilter_nkfilter_ntaken_nlength_0 nkfilter_nlength nlength_code) 
 have 2: "\<exists> x \<in> nset nell. P x"
   using assms(1) exists_nnth_nset by blast 
 have 3: "{i. i \<le> nlength(ntaken (nnth (nkfilter P 0 nell) k) nell) \<and> 
            P (nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) i) }
           = {i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> P (nnth nell i)}  "
   by (auto simp add: ntaken_nnth "1" order_subst2)      
 have 4: "\<exists> x \<in> nset(ntaken (nnth (nkfilter P 0 nell) k) nell). P x "
   using "1" assms(1) ntaken_nset by fastforce 
 have 5: "{i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> P (nnth nell i)} =
          {i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> i \<in> nset(nkfilter P 0 nell)}"
   using 4 1 2 nkfilter_holds_b
   by (metis (mono_tags, opaque_lifting) add_cancel_left_right enat_ord_simps(1) order.trans) 
 have 6: "{i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> i \<in> nset(nkfilter P 0 nell)} =
          {i. i\<le> (nnth (nkfilter P 0 nell) k) \<and> 
            i\<in> { (nnth (nkfilter P 0 nell) j) | j. j \<le> nlength (nkfilter P 0 nell)}}"
   by (simp add: nset_conv_nnth)
 have 7: "{i. i\<le> (nnth (nkfilter P 0 nell) k) \<and> 
            i\<in> { (nnth (nkfilter P 0 nell) j) | j. j \<le> nlength (nkfilter P 0 nell)}} =
          {(nnth (nkfilter P 0 nell) j) |j. j \<le> k} "
   by (auto simp add: assms "2" nidx_nkfilter_less_eq nkfilter_nlength)
      (metis "2" dual_order.antisym le_cases nidx_nkfilter_gr_eq nkfilter_nlength,
      metis assms(2) dual_order.trans enat_ord_simps(1))
 have 8: "k \<le> nlength (nkfilter P 0 nell)"
   by (simp add: "2" assms(2) nkfilter_nlength) 
 have 9: "{(nnth (nkfilter P 0 nell) j) |j. j \<le> k} = nset(ntaken k (nkfilter P 0 nell))"
   using ntaken_nset[of k "(nkfilter P 0 nell)"] 8 by auto
 have 91: "min (enat (nnth (nkfilter P 0 nell) k)) (nlength nell) = (nnth (nkfilter P 0 nell) k)"
   by (simp add: "1" min_def)
 have 10: "nset(nkfilter P 0 (ntaken ((nnth (nkfilter P 0 nell) k)) nell)) =
           {i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> 
               P (nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) i) } "
   using nkfilter_nset[of "(ntaken (nnth (nkfilter P 0 nell) k) nell)" P 0 ]
   1 4 91 ntaken_nlength[of "(nnth (nkfilter P 0 nell) k)" nell]
   by auto      
 have 11: "nlength((ntaken ((nnth (nkfilter P 0 nell) k)) nell)) = (nnth (nkfilter P 0 nell) k) "
   by (simp add: "1") 
 have 12: "{i. i \<le> (nnth (nkfilter P 0 nell) k) \<and> 
               P (nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) i) } =
           {i. i \<le> nlength((ntaken ((nnth (nkfilter P 0 nell) k)) nell)) \<and> 
               P (nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) i) }" 
   using 11 by auto
 show ?thesis 
 using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma ntaken_nkfilter_ntaken_nset_eq_1:
 assumes "\<exists> x \<in> nset nell. P x" 
         "k \<le> nlength (nfilter P nell) "
 shows   " nset(ntaken k (nkfilter P 0 nell)) =
           nset(nkfilter P 0 (ntaken ((nnth (nkfilter P 0 nell) k)) nell)) "
using assms ntaken_nkfilter_ntaken_nset_eq[of P nell k]
nkfilter_holds[of nell P "(nnth (nkfilter P 0 nell) k)" 0]  
by (metis in_nset_conv_nnth nkfilter_nlength nkfilter_nnth_n_zero) 
       
lemma ndropn_nkfilter_ndropn_nset_eq:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)) )" 
         "k \<le> nlength (nfilter P nell) "
 shows   " nset(ndropn k (nkfilter P 0 nell)) =
           nset(nkfilter P (nnth (nkfilter P 0 nell) k) (ndropn ((nnth (nkfilter P 0 nell) k)) nell)) "
proof -
 have 1: "(nnth (nkfilter P 0 nell) k) \<le> nlength nell"
   using nkfilter_upperbound[of nell P k 0]  assms
   by (metis diff_zero gen_nlength_def nfilter_nkfilter_ntaken_nlength_0 nkfilter_nlength nlength_code) 
 have 2: " \<exists> x \<in> nset nell. P x"
   using assms(1) exists_nnth_nset by blast 
 have 4: "\<exists> x \<in> nset(ndropn (nnth (nkfilter P 0 nell) k) nell). P x" 
   using "1" assms(1) ndropn_nset by fastforce 
 have 10: "nset(nkfilter P (nnth (nkfilter P 0 nell) k) 
                            (ndropn ((nnth (nkfilter P 0 nell) k)) nell)) =
           {(nnth (nkfilter P 0 nell) k) + i| i. i \<le> nlength nell - (nnth (nkfilter P 0 nell) k) \<and> 
               P (nnth nell (i+ (nnth (nkfilter P 0 nell) k)))  }" 
   using 1 
   nkfilter_ndropn_nset_b[of "(nnth (nkfilter P 0 nell) k)" nell P "(nnth (nkfilter P 0 nell) k)"]
   "4" by linarith 
 have 5: "{(nnth (nkfilter P 0 nell) k) + i| i. i \<le> nlength nell - (nnth (nkfilter P 0 nell) k) \<and> 
               P (nnth nell (i+ (nnth (nkfilter P 0 nell) k)))  } =
          {(nnth (nkfilter P 0 nell) k) + i| i. i \<le> nlength nell - (nnth (nkfilter P 0 nell) k) \<and> 
             (i+ (nnth (nkfilter P 0 nell) k)) \<in> nset(nkfilter P 0 nell)}"
    proof (auto simp add: add.commute) 
     fix i :: nat
     assume a1: "enat i \<le> nlength nell - enat (nnth (nkfilter P 0 nell) k)"
     assume a2: "P (nnth nell (i + nnth (nkfilter P 0 nell) k))"
     have f0: "enat (i + (nnth (nkfilter P 0 nell) k)) \<le> nlength nell" 
     using "1" a1 enat_min_eq by auto
     show "i + nnth (nkfilter P 0 nell) k \<in> nset (nkfilter P 0 nell)"
     by (metis Nat.add_0_right a2 f0 in_nset_conv_nnth nkfilter_holds_b) 
    next 
     fix i
     assume b0: "enat i \<le> nlength nell - enat (nnth (nkfilter P 0 nell) k) "
     assume b1: "i + nnth (nkfilter P 0 nell) k \<in> nset (nkfilter P 0 nell)"
     show " P (nnth nell (i + nnth (nkfilter P 0 nell) k)) "
     using nkfilter_holds[of nell P ] 2 by (metis b1 minus_nat.diff_0)
   qed
 have 51: "{(nnth (nkfilter P 0 nell) k) + i| i. i \<le> nlength nell - (nnth (nkfilter P 0 nell) k) \<and> 
             (i+ (nnth (nkfilter P 0 nell) k)) \<in> nset(nkfilter P 0 nell)} =
           {(nnth (nkfilter P 0 nell) k) + i| i. 
             (nnth (nkfilter P 0 nell) k) \<le> i + (nnth (nkfilter P 0 nell) k) \<and> 
             i + (nnth (nkfilter P 0 nell) k) \<le> nlength nell  \<and> 
             (i+ (nnth (nkfilter P 0 nell) k)) \<in> nset(nkfilter P 0 nell)}" 
   by auto
      (metis "2" add.left_neutral in_nset_conv_nnth nkfilter_upperbound zero_enat_def,
       metis add_diff_cancel_right' enat_minus_mono1 idiff_enat_enat)
 have 52: "{(nnth (nkfilter P 0 nell) k) + i| i. 
             (nnth (nkfilter P 0 nell) k) \<le> i + (nnth (nkfilter P 0 nell) k) \<and> 
             i + (nnth (nkfilter P 0 nell) k) \<le> nlength nell  \<and> 
             (i+ (nnth (nkfilter P 0 nell) k)) \<in> nset(nkfilter P 0 nell)} =
           {j. (nnth (nkfilter P 0 nell) k) \<le>j \<and> j \<le> nlength nell \<and> j \<in> nset(nkfilter P 0 nell)} " 
   by (metis (no_types, lifting) add.commute diff_add)
 have 53 : "{j. (nnth (nkfilter P 0 nell) k) \<le>j \<and> j \<le> nlength nell \<and> j \<in> nset(nkfilter P 0 nell)} =
            {j. (nnth (nkfilter P 0 nell) k) \<le>j \<and> j \<le> nlength nell \<and> j \<in> 
                {(nnth (nkfilter P 0 nell) jj) |jj. jj \<le> nlength (nkfilter P 0 nell)}}"
   by (simp add: nset_conv_nnth)
 have 54: "{j. (nnth (nkfilter P 0 nell) k) \<le>j \<and> j \<le> nlength nell \<and> j \<in> 
                {(nnth (nkfilter P 0 nell) jj) |jj. jj \<le> nlength (nkfilter P 0 nell)}} =
           {(nnth (nkfilter P 0 nell) j) |j. k \<le> j \<and> j \<le> nlength(nkfilter P 0 nell)} "
   using assms 2 
   by (auto,
       metis dual_order.antisym le_cases nidx_less_eq nidx_nkfilter nkfilter_nlength,
       meson nidx_nkfilter_gr_eq,
       metis gen_nlength_def nkfilter_upperbound nlength_code)      
 have 8: " k \<le> nlength (nkfilter P 0 nell)" 
    by (simp add: "2" assms(2) nkfilter_nlength) 
 have 9: "{(nnth (nkfilter P 0 nell) j) |j. k \<le> j \<and> j \<le> nlength(nkfilter P 0 nell)} =
          nset(ndropn k (nkfilter P 0 nell))"
   by (simp add: "8" ndropn_nset) 
 show ?thesis 
   using "10" "5" "51" "52" "53" "54" "9" by auto
qed

lemma ndropn_nkfilter_ndropn_nset_eq_1:
 assumes "\<exists> x \<in> nset nell. P x" 
         "k \<le> nlength (nfilter P nell) "
 shows   " nset(ndropn k (nkfilter P 0 nell)) =
           nset(nkfilter P (nnth (nkfilter P 0 nell) k) (ndropn ((nnth (nkfilter P 0 nell) k)) nell)) "
using assms ndropn_nkfilter_ndropn_nset_eq[of P nell k]
by (metis diff_zero in_nset_conv_nnth nkfilter_holds nkfilter_nlength) 
 
lemma nkfilter_nkfilter_ntaken:
 assumes "\<exists> x \<in> nset nell. P x" 
         "k \<le> nlength (nfilter P nell) "
 shows   " ntaken k (nkfilter P 0 nell) =
           (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell)) "
using assms 
by (simp add: nfilter_nkfilter_ntaken_nidx_a nfilter_nkfilter_ntaken_nidx_b_1 nidx_nset_eq 
    ntaken_nkfilter_ntaken_nset_eq_1)

lemma nkfilter_nkfilter_ndropn:
 assumes "\<exists> x \<in> nset nell. P x" 
         "k \<le> nlength (nfilter P nell) "
 shows   "ndropn k (nkfilter P 0 nell) =
          (nkfilter P (nnth (nkfilter P 0 nell) k) (ndropn (nnth (nkfilter P 0 nell) k) nell))"
proof -
 have 1: "P (nnth nell ( (nnth (nkfilter P 0 nell) k)) ) "
   using nkfilter_holds[of nell P ] assms
   by (metis diff_zero in_nset_conv_nnth nkfilter_nlength)
 show ?thesis
   by (metis "1" assms(1) assms(2) diff_zero ndropn_nkfilter_ndropn_nset_eq 
       nfilter_nkfilter_ndropn_nidx_a nfilter_nkfilter_ndropn_nidx_b_3 nidx_nset_eq) 
qed

lemma nkfilter_nmap_nfilter:
 assumes "\<exists> x \<in> nset nell. P x" 
 shows   "nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell) = nfilter P nell"
 using assms nellist_eq_nnth_eq[of "nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell)" "nfilter P nell"]
nkfilter_nfilter[of nell P _ 0]  nkfilter_nnth_n_zero[of nell P _ 0] 
by (simp add: nkfilter_nlength) 

lemma nfilter_nkfilter_ntaken:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)) )"
         "k \<le> nlength (nkfilter P 0 nell)"  
 shows   " ntaken k (nfilter P nell) = 
           nfilter P (ntaken (nnth (nkfilter P 0 nell) k) nell) "
proof -
 have 1: "\<exists> x \<in> nset nell. P x"
   using assms(1) exists_nnth_nset by blast
 have 2: " nfilter P nell = nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell) "
   using 1  assms by (simp add: nkfilter_nmap_nfilter)
 have 3: "ntaken k (nfilter P nell) = ntaken k (nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell))" 
   using 2 by simp
 have 4: "ntaken k (nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell)) =
          nmap (\<lambda>n. nnth nell n) (ntaken k (nkfilter P 0 nell))"
   by simp
 have 5: "\<exists> x \<in> nset (ntaken (nnth (nkfilter P 0 nell) k) nell). P x "
   using assms by (metis nfinite_ntaken nset_nlast ntaken_nlast) 
 have 6: "(nfilter P (ntaken (nnth (nkfilter P 0 nell) k) nell)) =
          nmap (\<lambda>s. nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) s) 
               (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell))  " 
   using 5 by (simp add: nkfilter_nmap_nfilter)
 have 7: "nmap (\<lambda>n. nnth nell n) (ntaken k (nkfilter P 0 nell)) =
          nmap (\<lambda>n. nnth nell n) (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell))"
   by (simp add: "1" "2" assms(2) nkfilter_nkfilter_ntaken) 
 have 70: "enat k \<le> nlength (nfilter P nell)"
   using "2" assms by auto
 have 71: "(\<And>z. z \<in> nset (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell)) \<Longrightarrow>
       (\<lambda>n. nnth nell n) z = (\<lambda>s. nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) s)  z)"
   using assms 1 70 ntaken_nkfilter_ntaken_nset_eq[of P nell k] 
        ntaken_nset[of k "(nkfilter P 0 nell)"] mem_Collect_eq
     nidx_nkfilter_gr_eq[of nell P _ _ 0]
   proof simp
     fix z :: nat
     assume a1: "enat k \<le> nlength (nkfilter P 0 nell)"
     assume a2: "\<And>k j. \<lbrakk>k \<le> j; enat j \<le> nlength (nkfilter P 0 nell)\<rbrakk> \<Longrightarrow> 
                        nnth (nkfilter P 0 nell) k \<le> nnth (nkfilter P 0 nell) j"
     assume a3: "z \<in> nset (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell))"
     assume "{nnth (nkfilter P 0 nell) i |i. i \<le> k} = 
              nset (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell))"
     then have "\<exists>n. z = nnth (nkfilter P 0 nell) n \<and> n \<le> k"
     using a3 by blast
     then have "z \<le> nnth (nkfilter P 0 nell) k"
     using a2 a1 by meson
     then show "nnth nell z = nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) z"
     by (simp add: ntaken_nnth)
   qed
 have 8: "nmap (\<lambda>n. nnth nell n) 
               (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell)) =
          nmap (\<lambda>s. nnth (ntaken (nnth (nkfilter P 0 nell) k) nell) s) 
               (nkfilter P 0 (ntaken (nnth (nkfilter P 0 nell) k) nell))"
   using 1 5 assms nellist.map_cong0
   using "71" by blast 
 show ?thesis 
 by (simp add: "3" "6" "7" "8")
qed

lemma nfilter_nkfilter_ntaken_1:
 assumes "\<exists> x \<in> nset nell. P x"
         "k \<le> nlength (nkfilter P 0 nell)"  
 shows   " ntaken k (nfilter P nell) = 
           nfilter P (ntaken (nnth (nkfilter P 0 nell) k) nell) "
using assms nfilter_nkfilter_ntaken[of P nell k]  
by (metis in_nset_conv_nnth nkfilter_holds nkfilter_nnth_n_zero)

lemma nkfilter_nmap_shift:
 assumes "\<exists> x \<in> nset nell. P x"
 shows " nmap (\<lambda>s. nnth nell (s+n)) (nkfilter P 0 nell) =
         nmap (\<lambda>s. nnth nell s) (nkfilter P n nell)"
proof -
 have 1: "nlength (nmap (\<lambda>s. nnth nell (s+n)) (nkfilter P 0 nell)) =
          nlength (nmap (\<lambda>s. nnth nell s) (nkfilter P n nell))"
   by (simp add: assms nkfilter_nlength) 
 have 2: "\<And>i. i\<le> nlength (nmap (\<lambda>s. nnth nell (s+n)) (nkfilter P 0 nell)) \<longrightarrow>
              nnth (nmap (\<lambda>s. nnth nell (s+n)) (nkfilter P 0 nell)) i =
              nnth nell ((nnth (nkfilter P 0 nell) i)+n)"
   by simp
 have 3: "\<And>i. i\<le> nlength (nmap (\<lambda>s. nnth nell s) (nkfilter P n nell)) \<longrightarrow>
              nnth (nmap (\<lambda>s. nnth nell s) (nkfilter P n nell)) i =
              nnth nell (nnth (nkfilter P n nell) i)"
   by simp
 have 4: " \<And>i. i\<le> nlength (nmap (\<lambda>s. nnth nell (s+n)) (nkfilter P 0 nell)) \<longrightarrow>
              nnth nell ((nnth (nkfilter P 0 nell) i)+n) = nnth nell (nnth (nkfilter P n nell) i)"
   using nkfilter_n_zero[of nell P n]
   by (simp add: assms)
 show ?thesis using "1" "4" nellist_eq_nnth_eq by force
qed

lemma nkfilter_nmap_shift_ndropn:
 assumes "\<exists> x \<in> nset (ndropn (nnth (nkfilter P 0 nell) k) nell). P x"
         "k \<le> nlength (nkfilter P 0 nell)" 
 shows " nmap (\<lambda>s. nnth nell (s+(nnth (nkfilter P 0 nell) k))) 
              (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k) nell)) =
         nmap (\<lambda>s. nnth nell s) (nkfilter P (nnth (nkfilter P 0 nell) k) 
                                            (ndropn (nnth (nkfilter P 0 nell) k) nell))"
using assms 
      nellist_eq_nnth_eq[of "nmap (\<lambda>s. nnth nell (s+(nnth (nkfilter P 0 nell) k))) 
                                  (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k) nell))" 
                            "nmap (\<lambda>s. nnth nell s) (nkfilter P (nnth (nkfilter P 0 nell) k) 
                                  (ndropn (nnth (nkfilter P 0 nell) k) nell))"]  
using nkfilter_n_zero[of "(ndropn (nnth (nkfilter P 0 nell) k) nell)" P 
                          "(nnth (nkfilter P 0 nell) k)"]
by simp 

lemma nfilter_nkfilter_ndropn:
 assumes "P (nnth nell ( (nnth (nkfilter P 0 nell) k)) )"
         "k \<le> nlength (nkfilter P 0 nell)"  
 shows   " ndropn k (nfilter P nell) = 
           nfilter P (ndropn (nnth (nkfilter P 0 nell) k) nell) "
proof -
 have 1: "\<exists>x \<in> nset nell. P x"
   using assms using exists_nnth_nset by blast 
 have 2: "(nfilter P nell) = nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell)" 
   by (simp add: "1" nkfilter_nmap_nfilter)
 have 3: "ndropn k (nfilter P nell) = ndropn k (nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell))"
   by (simp add: "2") 
 have 4: "ndropn k (nmap (\<lambda>n. nnth nell n) (nkfilter P 0 nell)) =
          nmap (\<lambda>n. nnth nell n) (ndropn k (nkfilter P 0 nell)) "
   using ndropn_nmap by blast
 have 5: "\<exists> x \<in> nset(ndropn (nnth (nkfilter P 0 nell) k) nell). P x"
   by (metis add.commute add.left_neutral assms(1) in_nset_conv_nnth ndropn_nnth zero_enat_def 
       zero_le) 
 have 6: "nfilter P (ndropn (nnth (nkfilter P 0 nell) k) nell) =
          nmap (\<lambda>s. nnth (ndropn (nnth (nkfilter P 0 nell) k) nell) s)
               (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k ) nell))" 
   by (metis "5" nkfilter_nmap_nfilter)
 have 7: "nmap (\<lambda>s. nnth (ndropn (nnth (nkfilter P 0 nell) k) nell) s)
               (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k ) nell)) =
          nmap (\<lambda>s. nnth nell (s+(nnth (nkfilter P 0 nell) k)))
               (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k ) nell))" 
   by (simp add: add.commute)
 have 8: "nmap (\<lambda>s. nnth nell (s+(nnth (nkfilter P 0 nell) k)))
               (nkfilter P 0 (ndropn (nnth (nkfilter P 0 nell) k ) nell)) =
          nmap (\<lambda>s. nnth nell s) 
               (nkfilter P (nnth (nkfilter P 0 nell) k) (ndropn (nnth (nkfilter P 0 nell) k) nell))"
   using "5" assms(2) nkfilter_nmap_shift_ndropn by fastforce 
 have 9: "nmap (\<lambda>s. nnth nell s) 
               (nkfilter P (nnth (nkfilter P 0 nell) k) (ndropn (nnth (nkfilter P 0 nell) k) nell)) =
          nmap (\<lambda>s. nnth nell s) 
               (ndropn k (nkfilter P 0 nell))"
   by (simp add: "1" "2" assms(2) nkfilter_nkfilter_ndropn) 
 show ?thesis using "3" "4" "6" "7" "8" "9" by auto
qed

lemma nfilter_nkfilter_ndropn_1:
 assumes "\<exists>x \<in> nset nell. P x"
         "k \<le> nlength (nkfilter P 0 nell)"  
 shows   " ndropn k (nfilter P nell) = 
           nfilter P (ndropn (nnth (nkfilter P 0 nell) k) nell) "
using assms nfilter_nkfilter_ndropn 
by (metis in_nset_conv_nnth minus_nat.diff_0 nkfilter_holds)

lemma nkfilter_nnth_aa:
 assumes "\<exists>x \<in> nset nell. P x"
         " n \<le> nlength (nfilter P nell)"
 shows   "P (nnth (nfilter P nell) n)"
using assms in_nset_conv_nnth nkfilter_holds[of nell P _ 0] nkfilter_nfilter[of nell P _ 0]
by (metis nkfilter_nlength)

lemma nfilter_nlength_zero_conv_a:
 assumes "\<exists>x \<in> nset nell. P x"
         "nlength(nfilter P nell) = 0"
 shows   "(\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j \<noteq>k \<longrightarrow> \<not> P (nnth nell j)))"
using assms 
apply transfer
proof (auto simp add: epred_llength min_def split: if_split_asm)
 fix nella :: "'a llist" and Pa :: "'a \<Rightarrow> bool" and x :: 'a
 assume a1: "\<not> lnull nella"
 assume a2: "Pa x"
 assume a3: "x \<in> lset nella"
 assume a4: "lnull (ltl (lfilter Pa nella))"
 show " \<exists>k. (enat k \<le> llength (ltl nella) \<longrightarrow>
            Pa (lnth nella k) \<and> (\<forall>j. enat j \<le> llength (ltl nella) \<longrightarrow> j \<noteq> k \<longrightarrow> \<not> Pa (lnth nella j))) \<and>
           enat k \<le> llength (ltl nella) " 
 proof -
  have 1: "LCons (lhd nella) (ltl nella) = nella"
    by (metis a2 a3 lfilter_LNil llist.disc(1) llist.exhaust_sel lnull_lfilter)
  have 2: "llength nella = eSuc (llength (ltl nella))"
    by (metis "1" llength_LCons)
  have 3: "\<not> lnull (lfilter Pa nella)"
    by (meson a2 a3 lnull_lfilter)  
  show ?thesis
    by (metis "2" "3" a4 iless_Suc_eq lfilter_llength_one_conv_a lhd_LCons_ltl llength_LCons llength_LNil 
        llist.collapse(1) one_eSuc) 
  qed   
qed

lemma nfilter_nlength_zero_conv_c:
 "(\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j \<noteq>k \<longrightarrow> \<not> P (nnth nell j))) =
  (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j<k \<or> k<j \<longrightarrow> \<not> P (nnth nell j)))"
using antisym_conv3 by  auto

lemma nfilter_nlength_zero_conv_d:
 " (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j<k \<or> k<j \<longrightarrow> \<not> P (nnth nell j))) \<longleftrightarrow>
   (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
          (\<forall>j. j<k  \<longrightarrow> \<not> P (nnth nell j)) \<and>
          (\<forall>j\<le> nlength nell. k<j \<longrightarrow> \<not> P (nnth nell j)))"
by (meson enat_ord_simps(2) le_cases le_less_trans)
    
lemma nfilter_nlength_zero_conv_b:
 assumes "(\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j \<noteq>k \<longrightarrow> \<not> P (nnth nell j)))"
 shows   " (\<exists>x \<in> nset nell. P x) \<and> nlength(nfilter P nell) = 0"
using assms 
by  transfer
    (auto split: if_split_asm,
     metis co.enat.exhaust_sel iless_Suc_eq in_lset_conv_lnth llength_eq_0,
     metis co.enat.exhaust_sel co.enat.sel(2) iless_Suc_eq lfilter_llength_one_conv_b llength_eq_0 
     one_eSuc)

lemma nfilter_nlength_zero_conv:
 "((\<exists>x \<in> nset nell. P x) \<and> nlength(nfilter P nell) = 0) \<longleftrightarrow>
   (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j \<noteq>k \<longrightarrow> \<not> P (nnth nell j)))"
using nfilter_nlength_zero_conv_a[of nell P]  nfilter_nlength_zero_conv_b[of nell P]
by blast

lemma nfilter_nlength_zero_conv_1:
 "((\<exists>x \<in> nset nell. P x) \<and> nlength(nfilter P nell) = 0) \<longleftrightarrow> 
  (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
            (\<forall>j\<le> nlength nell. j<k \<or> k<j \<longrightarrow> \<not> P (nnth nell j)))"
using nfilter_nlength_zero_conv[of nell P] nfilter_nlength_zero_conv_c[of nell P]
by blast

lemma nfilter_nlength_zero_conv_2:
 "((\<exists>x \<in> nset nell. P x) \<and> nlength(nfilter P nell) = 0) \<longleftrightarrow> 
   (\<exists> k \<le> nlength nell. P (nnth nell k) \<and> 
          (\<forall>j. j<k  \<longrightarrow> \<not> P (nnth nell j)) \<and>
          (\<forall>j\<le> nlength nell. k<j \<longrightarrow> \<not> P (nnth nell j))) "
using nfilter_nlength_zero_conv_1[of nell P] nfilter_nlength_zero_conv_d[of nell P]
by blast

lemma nfilter_ndropns_nmap_help_0:
 assumes "\<exists> x \<in> nset nell. P x"
         " j \<le> nnth(nkfilter P 0 nell) 0 "
 shows   "(nfilter P (ndropn (nnth (nkfilter P 0 nell) 0) nell)) = (nfilter P (ndropn j nell))"
 using assms 
  proof (induction j arbitrary: nell)
  case 0
  then show ?case
  by (metis ndropn_0 nfilter_nkfilter_ndropn_1 zero_enat_def zero_le) 
  next
  case (Suc j)
  then show ?case 
    proof (cases nell)
    case (NNil x1)
    then show ?thesis
    by simp 
    next
    case (NCons x21 x22)
    then show ?thesis 
       proof -
        have 1: "is_NNil x22 \<Longrightarrow> 
                 nfilter P (ndropn (nnth (nkfilter P 0 nell) 0) nell) = nfilter P (ndropn (Suc j) nell) "
          by (metis NCons Suc.prems(2) Suc_le_D ndropn_Suc_NCons ndropn_is_NNil)
        have 2: "P x21 \<and> \<not>is_NNil x22 \<Longrightarrow> 
                 nfilter P (ndropn (nnth (nkfilter P 0 nell) 0) nell) = nfilter P (ndropn (Suc j) nell)" 
          using Suc NCons 
          by simp
             (metis Suc.prems(1) dual_order.strict_trans1 nkfilter_not_before nnth_0 zero_less_Suc)
        have 3: "\<not>P x21 \<and> \<not>is_NNil x22 \<Longrightarrow> 
                 nfilter P (ndropn (nnth (nkfilter P 0 nell) 0) nell) = nfilter P (ndropn (Suc j) nell)" 
          using Suc NCons by (simp add: nkfilter_nleast)
        show ?thesis
          using "1" "2" "3" by blast   
      qed   
    qed
qed

lemma nfilter_nappend_ntaken:
 assumes "\<exists> x \<in> nset (ntaken k nell). P x"
         " k\<le> nlength nell"
 shows "nfilter P (ntaken k nell) = 
        ntaken (the_enat (nlength(nfilter P (ntaken k nell)))) (nfilter P nell)"
using assms
apply transfer
proof auto
 show " \<And>k nell P x.
        \<not> lnull nell \<Longrightarrow>
        enat k \<le> epred (llength nell) \<Longrightarrow>
        x \<in> lset (ltake (enat (Suc k)) nell) \<Longrightarrow>
        P x \<Longrightarrow>
        \<forall>x\<in>lset nell. \<not> P x \<Longrightarrow>
        lfilter P (ltake (enat (Suc k)) nell) =
        ltake (enat (Suc (the_enat (epred (llength (lfilter P (ltake (enat (Suc k)) nell))))))) nell" 
  by (metis dual_order.strict_trans1 in_lset_conv_lnth llength_ltake lprefix_lnthD ltake_is_lprefix 
    min.cobounded2)
next
 fix k 
 fix nell :: "'a llist" 
 fix P
 fix x
 fix xa
 assume a0: "\<not> lnull nell "
 assume a1: "enat k \<le> epred (llength nell) "
 assume a2: "x \<in> lset (ltake (enat (Suc k)) nell) "
 assume a3: "P x "
 assume a4: "xa \<in> lset nell "
 assume a5: "P xa "      
 show "lfilter P (ltake (enat (Suc k)) nell) =
       ltake (enat (Suc (the_enat (epred (llength (lfilter P (ltake (enat (Suc k)) nell))))))) (lfilter P nell)"
 proof (cases "k = epred (llength nell)")
 case True
 then show ?thesis
  proof -
  have f1: "eSuc (enat k) = llength nell"
  by (simp add: True a0)
  then have "llength (lfilter P nell) \<noteq> 0"
  by (metis (no_types) a2 a3 eSuc_enat llength_eq_0 lnull_lfilter ltake_all order_refl)
  then show ?thesis
   using f1 by (metis True co.enat.exhaust_sel eSuc_enat enat_the_enat epred_le_epredI infinity_ileE 
   llength_lfilter_ile ltake_all order_refl)
  qed
 next
 case False
 then show ?thesis
  proof -
  have f1: "llength nell \<noteq> 0"
  using a0 llength_eq_0 by blast
  have g1: "\<not> lnull (lfilter P (ltake (enat (Suc k)) nell))"
  by (meson a2 a3 lnull_lfilter)
  then show ?thesis
  using f1
  proof -
   have f1: "enat k < epred (llength nell)"
   using False a1 order.not_eq_order_implies_strict by blast
   have f2: "\<forall>e. e = 0 \<or> e = eSuc (epred e)"
   by (metis co.enat.exhaust_sel)
   then have g2: "enat (Suc k) < llength nell"
   using f1 by (metis (no_types) Suc_ile_eq \<open>llength nell \<noteq> 0\<close> iless_Suc_eq)
   then show ?thesis
   using f2
   using lfilter_lappend_ltake[of "Suc k"]
   proof -
    have "eSuc (enat k) = min (enat (Suc k)) (llength nell)"
    by (metis \<open>enat (Suc k) < llength nell\<close> eSuc_enat min.strict_order_iff)
    then have "eSuc (enat k) = llength (ltake (enat (Suc k)) nell)"
    by simp
    then show ?thesis
    by (metis g1 g2 co.enat.exhaust_sel eSuc_enat eSuc_infinity enat_the_enat infinity_ileE
        lfilter_lappend_ltake llength_eq_0 llength_lfilter_ile)
   qed 
  qed
  qed
 qed
qed

lemma nappend_ntaken_ndropn:
 assumes " (Suc k) \<le> nlength nell"
 shows   " nappend (ntaken k nell) (ndropn (Suc k) nell) = nell"
using assms
by transfer  (simp add: min_absorb1)

lemma nappend_nfilter_nfinite:
 assumes "\<exists> x \<in> nset (nellx). P x"
         "\<exists> x \<in> nset (nelly). P x"
         "nfinite nellx" 
 shows   " nfilter P (nappend nellx nelly) = nappend (nfilter P nellx) (nfilter P nelly) "
using assms
by transfer auto

lemma nfilter_nappend_ndropn:
 assumes "\<exists> x \<in> nset (ndropn (Suc k) nell). P x"
         "\<exists> x \<in> nset (ntaken k nell). P x" 
         " (Suc k) \<le> nlength nell"
 shows "nfilter P (ndropn (Suc k) nell) = 
        ndropn (the_enat (eSuc(nlength(nfilter P (ntaken k nell))))) (nfilter P nell)"
proof -
 have 1: "nfinite (nfilter P (ntaken k nell))" 
   by (metis Suc_ile_eq assms(3) dual_order.strict_implies_order enat_ile length_nfilter_le 
       min.absorb1 nfinite_conv_nlength_enat ntaken_nlength)
 have 2: "nfilter P nell = nappend (nfilter P (ntaken k nell)) (nfilter P (ndropn (Suc k) nell)) "
   using assms nappend_nfilter_nfinite[of "(ntaken k nell)" P "(ndropn (Suc k) nell)"] 
         nappend_ntaken_ndropn[of k nell]  nfinite_ntaken[of k nell] by simp
 have 3: "ndropn (the_enat (eSuc(nlength(nfilter P (ntaken k nell))))) 
                   (nappend (nfilter P (ntaken k nell)) (nfilter P (ndropn (Suc k) nell))) 
          =  (nfilter P (ndropn (Suc k) nell))" 
   by (metis "1" antisym_conv2 gr_zeroI ile_eSuc iless_Suc_eq less_imp_le ndropn_0 
        ndropn_nappend3 nfinite_conv_nlength_enat one_enat_def plus_1_eSuc(2) plus_enat_simps(1) 
        the_enat.simps zero_less_diff)
 show ?thesis 
   by (simp add: "2" "3")
qed

lemma nkfilter_nappend_ntaken:
 assumes "\<exists> x \<in> nset (ntaken k nell). P x"
         " k\<le> nlength nell"
 shows "nkfilter P n (ntaken k nell) = 
        ntaken (the_enat (nlength(nkfilter P n (ntaken k nell)))) (nkfilter P n nell)"
using assms
apply transfer
proof (auto simp add: kfilter_not_lnull_conv kfilter_lnull_conv)
 show "\<And>k nell P n x.
       \<not> lnull nell \<Longrightarrow>
       enat k \<le> epred (llength nell) \<Longrightarrow>
       x \<in> lset (ltake (enat (Suc k)) nell) \<Longrightarrow>
       P x \<Longrightarrow>
       \<forall>x\<in>lset nell. \<not> P x \<Longrightarrow>
       kfilter P n (ltake (enat (Suc k)) nell) =
       ltake (enat (Suc (the_enat (epred (llength (kfilter P n (ltake (enat (Suc k)) nell))))))) (iterates Suc n)" 
  by (meson lset_ltake subsetD)
 next
 fix ka :: nat and nella :: "'a llist" and Pa :: "'a \<Rightarrow> bool" and na :: nat and x :: 'a and xa :: 'a
 assume a1: "Pa x"
 assume a2: "Pa xa"
 assume a3: "xa \<in> lset nella"
 assume a4: "x \<in> lset (ltake (enat (Suc ka)) nella)"
 have f5: "\<And>e. e = enat 0 \<or> eSuc (epred e) = e"
 by (metis (no_types) co.enat.exhaust_sel zero_enat_def)
 have f6: "\<And>as. min \<infinity> (llength (as::'a llist)) = llength as"
 by simp
 have f7: "eSuc \<infinity> = \<infinity>"
 by simp
 have f8: "(enat (Suc (the_enat (epred (llength (kfilter Pa na (ltake (enat (Suc ka)) nella))))))) =
          (llength (kfilter Pa na (ltake (enat (Suc ka)) nella)))"
  proof -
   have f1: "\<forall>n. \<not> \<infinity> \<le> enat n"
   by (meson infinity_ileE)
   have "\<not> lnull (kfilter Pa na (ltake (enat (Suc ka)) nella))"
   by (metis (no_types) a1 a4 kfilter_not_lnull_conv)
   then show ?thesis
   using f1
   by (metis co.enat.exhaust_sel dual_order.trans eSuc_enat eSuc_ile_mono enat_the_enat kfilter_llength 
      llength_eq_0 llength_lfilter_ile llength_ltake min.cobounded1)
  qed
 moreover
 { assume "llength nella \<noteq> \<infinity>"
   then have "ltake (enat (Suc ka)) nella = nella \<longrightarrow> 
              ltake (llength (lfilter Pa (ltake (enat (Suc ka)) nella))) (kfilter Pa na nella) = 
              kfilter Pa na (ltake (enat (Suc ka)) nella) \<and>
               epred (llength (lfilter Pa (ltake (enat (Suc ka)) nella))) \<noteq> \<infinity>"
   using f7 f6 f5 a3 a2 by (metis (no_types) kfilter_llength llength_eq_0 llength_lfilter_ile 
    lnull_lfilter ltake_all min.orderE not_le_imp_less not_less_iff_gr_or_eq zero_enat_def)
  moreover
  { assume "ltake (enat (Suc ka)) nella \<noteq> nella"
    then have "enat (Suc ka) < llength nella \<and> min (enat (Suc ka)) (llength nella) \<noteq> \<infinity> \<or> 
               enat (Suc ka) < llength nella \<and> llength (lfilter Pa (ltake (enat (Suc ka)) nella)) \<noteq> eSuc \<infinity>"
    by (metis (no_types) enat_ord_code(4) ltake_all min.orderE not_le_imp_less not_less_iff_gr_or_eq) }
   ultimately have "enat (Suc ka) < llength nella \<and> min (enat (Suc ka)) (llength nella) \<noteq> \<infinity> \<or>  
                    enat (Suc ka) < llength nella \<and> llength (lfilter Pa (ltake (enat (Suc ka)) nella)) \<noteq> eSuc \<infinity> \<or>
                     ltake (llength (lfilter Pa (ltake (enat (Suc ka)) nella))) (kfilter Pa na nella) = 
                     kfilter Pa na (ltake (enat (Suc ka)) nella) \<and> 
                     epred (llength (lfilter Pa (ltake (enat (Suc ka)) nella))) \<noteq> \<infinity>"
   by fastforce }
  ultimately show "kfilter Pa na (ltake (enat (Suc ka)) nella) = 
                   ltake (enat (Suc (the_enat (epred (llength (kfilter Pa na (ltake (enat (Suc ka)) nella)))))))
                         (kfilter Pa na nella)"
  using f6 f5 a4 a1
  kfilter_lappend_ltake[of "(enat (Suc ka))" nella Pa na ]
  by (metis enat_ord_code(4) kfilter_llength)
qed

lemma nfilter_ndropns_nmap_help_1:
 assumes "\<exists> x \<in> nset nell. P x"
         " j \<le> nnth(nkfilter P 0 nell) (Suc 0) "
         " nnth (nkfilter P 0 nell) 0 < j" 
         " (Suc 0) \<le> nlength(nfilter P nell)"
 shows   "(nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc 0)) nell)) =
          (nfilter P (ndropn j nell))"
using assms
proof 
 (induction j arbitrary: nell )
 case 0
 then show ?case
 by blast 
 next
 case (Suc j)
 then show ?case 
     proof (cases nell)
     case (NNil x1)
     then show ?thesis
     by auto 
     next
     case (NCons x21 x22)
     then show ?thesis 
        proof -
         have 1: "is_NNil x22 \<Longrightarrow> 
                  nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc 0)) nell) = 
                  nfilter P (ndropn (Suc j) nell)"
           by (metis NCons Suc.prems(2) Suc_le_D ndropn_Suc_NCons ndropn_is_NNil) 
         have 2: "P x21 \<and> \<not> is_NNil x22 \<Longrightarrow>
                  nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc 0)) nell) = 
                  nfilter P (ndropn (Suc j) nell) " 
           using Suc NCons 
           proof simp
           assume a1: "P x21 \<and> \<not> is_NNil x22"
           assume a2: "enat (Suc 0) \<le> nlength (nfilter P (NCons x21 x22))"
           assume a3: "nell = NCons x21 x22"
           assume a4: "Suc j \<le> nnth (nkfilter P 0 (NCons x21 x22)) (Suc 0)"
           have f5: "\<forall>as p a n. ((\<exists>a. (a::'a) \<in> nset as \<and> p a) \<or> \<not> p a) \<or> 
                     nkfilter p n (NCons a as) = NNil n"
           by (metis (no_types) nkfilter_NCons_a)
           have f6: "\<forall>as p n. (\<forall>a. (a::'a) \<notin> nset as \<or> \<not> p a) \<or> 
                       nlength (nkfilter p n as) = nlength (nfilter p as)"
           by (meson nkfilter_nlength)
           have f7: "\<forall>p as. (\<forall>a. (a::'a) \<notin> nset as \<or> \<not> p a) = (\<forall>a. a \<notin> nset as \<or> \<not> p a)"
           by meson
           have f8: "nlength (nkfilter P 0 (NCons x21 x22)) = 
                          nlength (nfilter P (NCons x21 x22))"
           by (meson a1 nellist.set_intros(2) nkfilter_nlength)
           have f9: "\<forall>p as. (\<forall>a. (a::'a) \<notin> nset as \<or> \<not> p a) = (\<forall>a. a \<notin> nset as \<or> \<not> p a)"
           by meson
           have f10: "(\<exists>a. a \<in> nset x22 \<and> P a) \<longrightarrow> 
                       nkfilter P 0 (NCons x21 x22) = NCons 0 (nkfilter P (Suc 0) x22)"
           using a1 by (simp add: Bex_def_raw)
           then have f11: "(\<exists>a. a \<in> nset x22 \<and> P a) \<longrightarrow> 
                            nnth (nkfilter P (Suc 0) x22) 0 - Suc 0 = nnth (nkfilter P 0 x22) 0"
           using f9 f8 f7 a2 by (metis Suc_ile_eq iless_Suc_eq nkfilter_nnth_n_zero nlength_NCons)
           have f14: "j < nnth (nkfilter P 0 (NCons x21 x22)) (Suc 0)"
           using a4 Suc_le_eq by blast
           have "\<exists>a. a \<in> nset x22 \<and> P a"
           using f8 f5 a2 a1 by (metis Suc_ile_eq gr_implies_not_zero nlength_NNil)
           then have "(\<exists>a. a \<in> nset x22 \<and> P a) \<and> nfilter P (ndropn (nnth (nkfilter P 0 x22) 0) x22) = 
                         nfilter P (ndropn j x22)"
           using f14   f11 a4
           proof -
           have "j \<le> nnth (nkfilter P 0 x22) 0"
           using \<open>\<exists>a. a \<in> nset x22 \<and> P a\<close> f10 f11 f14 by fastforce
           then show ?thesis
           by (simp add: Bex_def_raw \<open>\<exists>a. a \<in> nset x22 \<and> P a\<close> nfilter_ndropns_nmap_help_0)
           qed  
           then show "nfilter P (ndropn (nnth (nkfilter P 0 (NCons x21 x22)) (Suc 0)) (NCons x21 x22)) = 
                      nfilter P (ndropn j x22)"
           using   f11 a4
           by (metis One_nat_def Suc_le_D diff_Suc_1 f10 ndropn_Suc_NCons nnth_Suc_NCons)
        qed 
      have 3: "\<not>P x21 \<and> \<not> is_NNil x22 \<Longrightarrow>
                  nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc 0)) nell) = 
                  nfilter P (ndropn (Suc j) nell) " 
        using Suc NCons
        proof simp
           assume a1: "\<And>nell. \<lbrakk>\<exists>a\<in>nset nell. P a; j \<le> nnth (nkfilter P 0 nell) (Suc 0); 
                                nnth (nkfilter P 0 nell) 0 < j; enat (Suc 0) \<le> nlength (nfilter P nell)\<rbrakk> \<Longrightarrow> 
                               nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc 0)) nell) = 
                               nfilter P (ndropn j nell)"
           assume a2: "enat (Suc 0) \<le> nlength (nfilter P x22)"
           assume a3: "Suc j \<le> nnth (nkfilter P (Suc 0) x22) (Suc 0)"
           assume a4: "\<exists>x\<in>nset x22. P x"
           assume a5: "nnth (nkfilter P (Suc 0) x22) 0 < Suc j"
           have f12: "nnth (nkfilter P 0 x22) (Suc 0) = 
                           ( (nnth (nkfilter P (Suc 0) x22) (Suc 0)) - (Suc 0))"
           using nkfilter_nnth_n_zero[of x22 P "Suc 0" "Suc 0" ]
           by (simp add: a2 a4 nkfilter_nlength)  
           then have f13: "j \<le> nnth (nkfilter P 0 x22) (Suc 0)"
           using a3 by linarith
           have f14: "enat 0 \<le> nlength (nfilter P x22)"
           using a2 by (metis One_nat_def one_enat_def order.trans zero_enat_def zero_le_one)
           have f15: "nlength (nkfilter P (Suc 0) x22) = nlength (nfilter P x22)"
           by (simp add: a4 nkfilter_nlength)
           then have "Suc 0 \<le> nnth (nkfilter P (Suc 0) x22) 0"
           by (simp add: a4 f14 nkfilter_lowerbound)
           then have "Suc (nnth (nkfilter P 0 x22) 0) \<le> j"
           by (metis a4 a5 add_Suc leD nkfilter_nleast not_less_eq_eq)
           then have "nfilter P (ndropn (nnth (nkfilter P 0 x22) (Suc 0)) x22) = 
                      nfilter P (ndropn j x22)"
           by (simp add: Suc.IH Suc_le_lessD a2 a4 f13)
           then show "nfilter P (ndropn (nnth (nkfilter P (Suc 0) x22) (Suc 0)) (NCons x21 x22)) = 
                      nfilter P (ndropn j x22)"
           using ndropn_Suc_NCons 
           by (metis One_nat_def a2 a4 f12 f15 le_add_diff_inverse nkfilter_lowerbound plus_1_eq_Suc)
         qed
       show ?thesis
         using "1" "2" "3" by blast   
     qed
 qed
qed

lemma nfilter_ndropns_nmap_help_j:
 assumes "\<exists> x \<in> nset nell. P x"
         " j \<le> nnth(nkfilter P 0 nell) (Suc i) "
         " nnth (nkfilter P 0 nell) i < j" 
         " (Suc i) \<le> nlength(nfilter P nell)"
 shows   "(nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell)) =
          (nfilter P (ndropn j nell))"
using assms
proof (induction j arbitrary: nell i)
case 0
then show ?case
by blast 
next
case (Suc j)
then show ?case 
  proof (cases nell)
  case (NNil x1)
  then show ?thesis
  by simp 
  next
  case (NCons x21 x22)
  then show ?thesis 
     proof -
      have 1: "is_NNil x22 \<Longrightarrow> 
                nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) = 
                nfilter P (ndropn (Suc j) nell) "
        by (metis NCons Suc.prems(2) Suc_le_D ndropn_Suc_NCons ndropn_is_NNil)
      have 2: "\<not> is_NNil x22 \<and> i = 0 \<Longrightarrow>
                nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) = 
                nfilter P (ndropn (Suc j) nell) "
        using Suc nfilter_ndropns_nmap_help_1[of nell P ]  by metis
      have 3: "P x21 \<and> \<not> is_NNil x22 \<and> 0< i \<Longrightarrow>
                nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) = nfilter P (ndropn (Suc j) nell)" 
        using Suc NCons
        proof simp
        assume a1: "\<And>nell i. \<lbrakk>\<exists>a\<in>nset nell. P a; j \<le> nnth (nkfilter P 0 nell) (Suc i); 
                              nnth (nkfilter P 0 nell) i < j; enat (Suc i) \<le> nlength (nfilter P nell)\<rbrakk> \<Longrightarrow> 
                        nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) = 
                        nfilter P (ndropn j nell)"
        assume a2: "P x21 \<and> \<not> is_NNil x22 \<and> 0 < i"
        assume a3: "enat (Suc i) \<le> nlength (nfilter P (NCons x21 x22))"
        assume a4: "nell = NCons x21 x22"
        assume a5: "Suc j \<le> nnth (nkfilter P 0 (NCons x21 x22)) (Suc i)"
        assume a6: "nnth (nkfilter P 0 (NCons x21 x22)) i < Suc j"
        show "nfilter P (ndropn (nnth (nkfilter P 0 (NCons x21 x22)) (Suc i)) (NCons x21 x22)) = 
                   nfilter P (ndropn j x22)"
        proof -
         have f0: "\<exists>a\<in>nset x22. P a"
           by (metis a2 a3 dual_order.antisym enat.inject nat.distinct(1) nfilter_NCons_a 
               nlength_NNil zero_enat_def zero_le) 
         have f1: "nlength (nkfilter P 0 (NCons x21 x22)) = nlength (nfilter P (NCons x21 x22))"
           using a4 by (metis (no_types) Suc(2) nkfilter_nlength)
         have f2: "nkfilter P 0 (NCons x21 x22) = NCons 0 (nkfilter P (Suc 0) x22)"
           by (metis a2 a3 dual_order.antisym enat.inject f1 nat.distinct(1) nkfilter_NCons 
               nkfilter_NCons_a nlength_NNil zero_enat_def zero_le)
         have f3: "nnth (nkfilter P 0 (NCons x21 x22)) (Suc i) =
                   nnth ( (nkfilter P (Suc 0) x22)) (i)"
           by (simp add: f2)
         have f4: "enat i \<le> nlength (nkfilter P (Suc 0) x22) "
           by (metis Suc_ile_eq a3 f1 f2 iless_Suc_eq nlength_NCons)
         have f5: "nnth ( (nkfilter P (Suc 0) x22)) (i) = (Suc 0) +nnth ( (nkfilter P 0 x22)) (i)" 
           using nkfilter_nnth_n_zero[of x22 P i "Suc 0" ] 
           using a5 f0 f3 f4 by linarith
         have f6: "ndropn (nnth (nkfilter P 0 (NCons x21 x22)) (Suc i)) (NCons x21 x22) =
                   ndropn ((Suc 0) +nnth ( (nkfilter P 0 x22)) (i)) (NCons x21 x22)" 
           using f3 f5 by presburger
         have f7: "ndropn ((Suc 0) +nnth ( (nkfilter P 0 x22)) (i)) (NCons x21 x22) =
                   ndropn (nnth ( (nkfilter P 0 x22)) (i)) x22 " 
           using ndropn_Suc_NCons by auto
         have f8: "j \<le> nnth (nkfilter P 0 x22) (i)"
           using a5 f3 f5 by linarith 
         have f11: "nnth (nkfilter P 0 (NCons x21 x22)) i  = nnth ( (nkfilter P (Suc 0) x22)) (i-1)"
           by (metis One_nat_def Suc_pred a2 f2 nnth_Suc_NCons)
         have f12: "enat (i - 1) \<le> nlength (nkfilter P (Suc 0) x22)"
           by (metis One_nat_def Suc_ile_eq Suc_pred a2 f4 less_imp_le)
         have f13: "(Suc 0) \<le> nnth ( (nkfilter P (Suc 0) x22)) (i-1)"
           by (meson f0 f12 nkfilter_lowerbound) 
         have f14: "nnth ( (nkfilter P (Suc 0) x22)) (i-1) = (Suc 0) +nnth ( (nkfilter P 0 x22)) (i-1)"
           using nkfilter_nnth_n_zero[of x22 P "i-1" "Suc 0" ] f12 f0  f13 by (metis le_add_diff_inverse)
         have f9: "nnth (nkfilter P 0 x22) (i-1) < j" 
           using a6 f11 f14 by linarith
         have f10: " i \<le> nlength (nfilter P x22)" 
           by (metis f0 f4 nkfilter_nlength)
         show ?thesis using a1[of x22 "i-1"] 
           by (metis Suc_diff_1 a2 f0 f10 f3 f5 f7 f8 f9) 
       qed
      qed
      have 4: "\<not>P x21 \<and> \<not> is_NNil x22 \<and> 0< i \<Longrightarrow>
                nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) = 
                nfilter P (ndropn (Suc j) nell)" 
        using Suc NCons 
        proof simp
         assume a0: "\<not> P x21 \<and> \<not> is_NNil x22 \<and> 0 < i " 
         assume a2: "\<And>nell i. \<lbrakk>\<exists>a\<in>nset nell. P a; j \<le> nnth (nkfilter P 0 nell) (Suc i); 
                             nnth (nkfilter P 0 nell) i < j; enat (Suc i) \<le> nlength (nfilter P nell)\<rbrakk> \<Longrightarrow>
                      nfilter P (ndropn (nnth (nkfilter P 0 nell) (Suc i)) nell) =
                      nfilter P (ndropn j nell)"
         assume a1: "\<exists>x\<in>nset x22. P x"
         assume a4: "Suc j \<le> nnth (nkfilter P (Suc 0) x22) (Suc i)"
         assume a5: "nnth (nkfilter P (Suc 0) x22) i < Suc j"
         assume a3: "enat (Suc i) \<le> nlength (nfilter P x22)"
         assume a6: "nell = NCons x21 x22" 
         show "nfilter P (ndropn (nnth (nkfilter P (Suc 0) x22) (Suc i)) (NCons x21 x22)) = 
              nfilter P (ndropn j x22)"
         proof - 
          have f1: "nlength (nkfilter P 0 (NCons x21 x22)) = nlength (nfilter P (NCons x21 x22))"
          by (metis NCons Suc.prems(1) nkfilter_nlength)
          have f2: "nkfilter P 0 (NCons x21 x22) =  (nkfilter P (Suc 0) x22)" 
          using NCons Suc.prems(2) a1 a5 by auto
          have f3: "nnth (nkfilter P 0 (NCons x21 x22)) (Suc i) =
                  nnth (nkfilter P (Suc 0) x22) (Suc i)" 
          by (simp add: f2)
          have f4: "enat (Suc i) \<le> nlength (nkfilter P (Suc 0) x22)" 
          using NCons Suc.prems(4) f1 f2 by auto
          have f5: "nnth ( (nkfilter P (Suc 0) x22)) (Suc i) = (Suc 0) +nnth ( (nkfilter P 0 x22)) (Suc i)" 
          by (metis One_nat_def Suc_le_D a1 a4 diff_Suc_1 f4 nkfilter_nnth_n_zero plus_1_eq_Suc)
          have f6: "(ndropn (nnth (nkfilter P (Suc 0) x22) (Suc i)) (NCons x21 x22)) =
                   ndropn ( (Suc 0) +nnth ( (nkfilter P 0 x22)) (Suc i)) (NCons x21 x22)"
          by (simp add: f5) 
          have f7: "ndropn ( (Suc 0) +nnth ( (nkfilter P 0 x22)) (Suc i)) (NCons x21 x22) =
                  ndropn ( nnth ( (nkfilter P 0 x22)) (Suc i)) (x22)" 
          by simp
          have f8: "j \<le> nnth (nkfilter P 0 x22) (Suc i)"
          using a4 f5 by force
          have f11: "nnth (nkfilter P 0 (NCons x21 x22)) i  = nnth ( (nkfilter P (Suc 0) x22)) (i)"
          by (simp add: f2)
          have f12: "enat (i ) \<le> nlength (nkfilter P (Suc 0) x22)"
          using Suc_ile_eq f4 by auto  
          have f13: "(Suc 0) \<le> nnth ( (nkfilter P (Suc 0) x22)) (i)"
          by (simp add: a1 f12 nkfilter_lowerbound)
          have f14: "nnth ( (nkfilter P (Suc 0) x22)) (i) = (Suc 0) +nnth ( (nkfilter P 0 x22)) (i)"
          by (metis a1 f12 f13 le_add_diff_inverse nkfilter_nnth_n_zero)
          have f9: "nnth (nkfilter P 0 x22) (i) < j" 
          using a5 f14 by auto 
          show ?thesis
          using Suc.IH a1 a3 f6 f7 f8 f9 by presburger
      qed
     qed
     show ?thesis
     using "1" "2" "3" "4" by fastforce 
    qed
  qed
qed

lemma nfilter_ndropns_nmap:
 assumes "\<exists> x \<in> nset (ndropns nell). P x"
         "(Suc i) \<le> nlength(nfilter P (ndropns nell))"
 shows " ndropn (Suc i) (nmap (\<lambda>s. nnth s 0) (nfilter P (ndropns nell))) =
         (nmap (\<lambda>s. nnth s 0) 
               (nfilter P (ndropns (ndropn (Suc (nnth (nkfilter P 0 (ndropns nell)) i)) nell)))) "
proof -
 have 1: "ndropn (Suc i) (nmap (\<lambda>s. nnth s 0) (nfilter P (ndropns nell))) =
          nmap (\<lambda>s. nnth s 0) (ndropn (Suc i) (nfilter P (ndropns nell)))"
   by (simp add: ndropn_nmap) 
 have 2: "(Suc (nnth (nkfilter P 0 (ndropns nell)) i)) \<le> nlength (ndropns nell)"
   using assms nkfilter_nkfilter_ntaken[of "(ndropns nell)" P  i] 
   by (metis Suc_ile_eq antisym_conv2 less_imp_le min.orderE nkfilter_nlength not_le_imp_less 
       ntaken_all ntaken_nlength)
 have 3: "(nfilter P (ndropns (ndropn (Suc (nnth (nkfilter P 0 (ndropns nell)) i)) nell))) =
          (nfilter P (ndropn (Suc (nnth (nkfilter P 0 (ndropns nell)) i)) (ndropns nell))) "
   by (simp add: "2" ndropn_ndropns) 
 have 4: "(ndropn (Suc i) (nfilter P (ndropns nell))) =
          (nfilter P (ndropn (nnth (nkfilter P 0 (ndropns nell)) (Suc i)) (ndropns nell))) "
   by (simp add: assms(1) assms(2) nfilter_nkfilter_ndropn_1 nkfilter_nlength)
 have 5: " (Suc (nnth (nkfilter P 0 (ndropns nell)) i)) \<le> (nnth (nkfilter P 0 (ndropns nell)) (Suc i)) " 
   by (simp add: Suc_leI assms(1) assms(2) nidx_nkfilter_expand nkfilter_nlength)
 have 6: " i < nlength(nfilter P (ndropns nell))"
   using Suc_ile_eq assms(2) by blast 
 have 7: "nnth (nkfilter P 0 (ndropns nell)) i < (Suc (nnth (nkfilter P 0 (ndropns nell)) i))"
   by simp
 have 8: "(nfilter P (ndropn (nnth (nkfilter P 0 (ndropns nell)) (Suc i)) (ndropns nell))) =
          (nfilter P (ndropn (Suc (nnth (nkfilter P 0 (ndropns nell)) i)) (ndropns nell)))"
   using 5 6 7 assms
   nfilter_ndropns_nmap_help_j[of "ndropns nell" P "(Suc (nnth (nkfilter P 0 (ndropns nell)) i))" i ]
   by blast   
 show ?thesis 
   using "1" "3" "4" "8" by auto
qed

lemma ndropns_nfilter_nnth_exist_ndropn:
 assumes "\<exists> x \<in> nset (ndropns nell). P x"
         " j \<le> nlength (nfilter P (ndropns nell)) "
 shows " (\<exists>i. enat i \<le> nlength nell \<and> nnth (nfilter P (ndropns nell)) j = ndropn i nell)"
proof -
 have 1: "nnth (nfilter P (ndropns nell)) j \<in> nset (ndropns nell)" 
   using assms in_nset_conv_nnth nfilter_nnth by fastforce
 show ?thesis using "1"using in_nset_ndropns by blast
qed

lemma nfilter_ndropns_nnth_bound:
 assumes "(\<exists> ys \<in> nset (ndropns xs). P ys) "
         " j \<le> nlength (nfilter P (ndropns xs))"
 shows  " nlength ((nnth (nfilter P (ndropns xs)) j)) \<le> nlength xs"
using assms ndropns_nfilter_nnth_exist_ndropn[of xs P j] 
by (metis add.commute enat.simps(3) enat_add_sub_same enat_le_plus_same(1) less_eqE ndropn_nlength) 

lemma ndropns_nfilter_ndropn:
 assumes "(Suc k) \<le> nlength nell"
         "\<exists> x \<in> nset (ndropns (ndropn (Suc k) nell)). P x"
         "\<exists>x\<in>nset (ntaken k (ndropns nell)). P x" 
 shows  "(nfilter P (ndropns (ndropn (Suc k) nell))) =  
         (ndropn (the_enat (eSuc(nlength(nfilter P (ntaken k (ndropns nell)))))) (nfilter P (ndropns nell))) "
using assms 
ndropn_ndropns[of "Suc k" nell] ndropns_nlength[of nell] nfilter_nappend_ndropn[of k "ndropns nell" P]    
by simp

lemma ndropns_nfilter_ndropn_a:
 assumes " k \<le> nlength(nfilter P (ndropns nell))"
         " \<exists> x \<in> nset (ndropns nell). P x"
 shows   " ndropn k (nfilter P (ndropns nell)) =
           nfilter P (ndropns (ndropn (nnth (nkfilter P 0 (ndropns nell)) k) nell)) "
using assms ndropn_ndropns   nfilter_nkfilter_ndropn_1[of "ndropns nell" P k] 
 nkfilter_upperbound[of "ndropns nell" P k 0] 
by (metis (mono_tags, lifting) add.left_neutral nkfilter_nlength zero_enat_def) 

lemma nfilter_nlength_imp:
 assumes " \<exists> x \<in> nset xs. P x \<and> Q x"
 shows   " nlength (nfilter (\<lambda>x. P x \<and> Q x) xs) \<le> nlength (nfilter P xs)"
using assms by transfer (auto simp add: lfilter_nlength_imp epred_le_epredI)

lemma nkfilter_chop: 
assumes "nlast xs = nfirst ys"
        " P (nlast xs)"
        "nfinite xs" 
shows "nkfilter P k (nfuse xs ys) = 
       nfuse (nkfilter P k xs)  (nkfilter P (k+(the_enat (nlength xs))) ys)"
using assms 
proof (auto simp add: nfuse_def)
show "nlast xs = nfirst ys \<Longrightarrow>
    P (nfirst ys) \<Longrightarrow>
    nfinite xs \<Longrightarrow>
    is_NNil ys \<Longrightarrow>
    \<not> is_NNil (nkfilter P (k + the_enat (nlength xs)) ys) \<Longrightarrow> 
    nkfilter P k xs = nappend (nkfilter P k xs) (ntl (nkfilter P (k + the_enat (nlength xs)) ys))"
  by transfer  auto
show "nlast xs = nfirst ys \<Longrightarrow>
    P (nfirst ys) \<Longrightarrow>
    nfinite xs \<Longrightarrow> \<not> is_NNil ys \<Longrightarrow> 
    is_NNil (nkfilter P (k + the_enat (nlength xs)) ys) \<Longrightarrow> 
    nkfilter P k (nappend xs (ntl ys)) = nkfilter P k xs"  
  apply transfer 
  proof (auto simp add: kfilter_lappend_lfinite lappend_lnull2)
  fix xsa :: "'a llist" and ysa :: "'a llist" and Pa :: "'a \<Rightarrow> bool" and ka :: nat and b :: nat
  assume a1: "(if lnull (kfilter Pa (ka + the_enat (epred (llength xsa))) ysa) 
               then iterates Suc (ka + the_enat (epred (llength xsa)))
               else kfilter Pa (ka + the_enat (epred (llength xsa))) ysa) = LCons b LNil"
  assume a2: "\<not> lnull ysa"
  assume a3: "Pa (lhd ysa)"
  assume a4: "\<not> lnull (kfilter Pa (ka + the_enat (llength xsa)) (ltl ysa))"
  have f5: "lnull (LNil::nat llist)"
  using llength_LNil llength_eq_0 by blast
  have f6: "ysa = LCons (lhd ysa) (ltl ysa)"
   using a2 by auto 
  then have "Pa (lhd ysa)"
    using a3 by auto
  then have False
   by (metis a1 a2 a4 eq_LConsD f6 kfilter_code(2) kfilter_lnull_conv llength_LNil llength_eq_0 llist.set_sel(1))
  then show "lappend (kfilter Pa ka xsa) (kfilter Pa (ka + the_enat (llength xsa)) (ltl ysa)) = iterates Suc ka"
  by meson
  next
   fix xsa :: "'b llist" and ysa :: "'b llist" and Pa :: "'b \<Rightarrow> bool" and ka :: nat and b :: nat
   assume a1: "(if lnull (kfilter Pa (ka + the_enat (epred (llength xsa))) ysa) 
                then iterates Suc (ka + the_enat (epred (llength xsa)))
                else kfilter Pa (ka + the_enat (epred (llength xsa))) ysa) = LCons b LNil"
   assume a2: "\<not> lnull ysa"
   assume a3: "Pa (lhd ysa)"
   assume a4: "\<not> lnull (kfilter Pa (ka + the_enat (llength xsa)) (ltl ysa))"
   have f5: "lnull (LNil::nat llist)"
   using llength_LNil llength_eq_0 by blast
   have f6: "ysa = LCons (lhd ysa) (ltl ysa)"
   using a2 by auto
   then have "Pa (lhd ysa)"
   using a3 by auto
   then have False
   using f6 f5 a4 a1 by (metis kfilter_LCons kfilter_llength_n_zero llength_eq_0 ltl_simps(2) not_lnull_conv)
   then show "lappend (kfilter Pa ka xsa) (kfilter Pa (ka + the_enat (llength xsa)) (ltl ysa)) = 
                  kfilter Pa ka xsa"
    by meson
   qed
next
show "nlast xs = nfirst ys \<Longrightarrow>
    P (nfirst ys) \<Longrightarrow>
    nfinite xs \<Longrightarrow>
    \<not> is_NNil ys \<Longrightarrow>
    \<not> is_NNil (nkfilter P (k + the_enat (nlength xs)) ys) \<Longrightarrow>
    nkfilter P k (nappend xs (ntl ys)) = 
    nappend (nkfilter P k xs) (ntl (nkfilter P (k + the_enat (nlength xs)) ys)) " 
    apply transfer
    proof (auto simp add: lappend_iterates kfilter_lnull_conv split:if_split_asm)
     show f1:"\<And>xs ys P k x xa.
       \<not> lnull xs \<Longrightarrow>
       \<not> lnull ys \<Longrightarrow>
       llast xs = lhd ys \<Longrightarrow>
       P (lhd ys) \<Longrightarrow>
       lfinite xs \<Longrightarrow>
       \<forall>b. ys \<noteq> LCons b LNil \<Longrightarrow>
       \<forall>b. kfilter P (k + the_enat (epred (llength xs))) ys \<noteq> LCons b LNil \<Longrightarrow>
       x \<in> lset ys \<Longrightarrow> P x \<Longrightarrow> \<forall>x\<in>lset xs. \<not> P x \<Longrightarrow> P xa \<Longrightarrow> xa \<in> lset (ltl ys) \<Longrightarrow>
          kfilter P k (lappend xs (ltl ys)) = iterates Suc k" 
    by (metis in_lset_lappend_iff lappend_lbutlast_llast_id lbutlast_lfinite llist.set_intros(1))
    next
     show f2:"\<And>xs ys P k x xa xb.
       \<not> lnull xs \<Longrightarrow>
       \<not> lnull ys \<Longrightarrow>
       llast xs = lhd ys \<Longrightarrow>
       P (lhd ys) \<Longrightarrow>
       lfinite xs \<Longrightarrow>
       \<forall>b. ys \<noteq> LCons b LNil \<Longrightarrow>
       \<forall>b. kfilter P (k + the_enat (epred (llength xs))) ys \<noteq> LCons b LNil \<Longrightarrow>
       x \<in> lset ys \<Longrightarrow>
       P x \<Longrightarrow>
       xa \<in> lset xs \<Longrightarrow>
       P xa \<Longrightarrow>
       P xb \<Longrightarrow>
       xb \<in> lset xs \<Longrightarrow> 
       kfilter P k (lappend xs (ltl ys)) = 
       lappend (kfilter P k xs) (ltl (kfilter P (k + the_enat (epred (llength xs))) ys))"
       proof -
       fix xsa :: "'b llist" and ysa :: "'b llist" and Pa :: "'b \<Rightarrow> bool" and
           ka :: nat and x :: 'b and xa :: 'b and xb :: 'b
       assume a1: "\<not> lnull ysa"
       assume a2: "Pa (lhd ysa)"
       assume a3: "llast xsa = lhd ysa"
       assume a4: "\<not> lnull xsa"
       assume a5: "lfinite xsa"
       have f6: "ysa = LCons (lhd ysa) (ltl ysa)"
       using a1 by auto
       have f7: "LCons (lhd ysa) (ltl ysa) \<noteq> LNil \<and> lhd (LCons (lhd ysa) (ltl ysa)) = lhd ysa \<and> 
                 ltl (LCons (lhd ysa) (ltl ysa)) = ltl ysa"
       by auto
       then have f8: "kfilter Pa (the_enat (epred (llength xsa)) + ka) (LCons (lhd ysa) (ltl ysa)) =
                      LCons (the_enat (epred (llength xsa)) + ka) 
                            (kfilter Pa (Suc (the_enat (epred (llength xsa)) + ka)) (ltl ysa))"
       using f6 a2 by (metis (no_types) kfilter_LCons)
       have f9: "\<forall>bs p n bsa. \<not> lfinite (bs::'b llist) \<or> kfilter p n (lappend bs bsa) = 
                  lappend (kfilter p n bs) (kfilter p (n + the_enat (llength bs)) bsa)"
       by (meson kfilter_lappend_lfinite)
       have f10: "lfinite (lbutlast xsa)"
       using a5 by (meson lbutlast_lfinite)
       have f11: "lappend (kfilter Pa ka (lbutlast xsa)) 
                          (kfilter Pa (ka + the_enat (llength (lbutlast xsa))) LNil) = 
                  kfilter Pa ka (lbutlast xsa)"
       by simp
       have "LCons (the_enat (epred (llength xsa)) + ka) LNil = 
             kfilter Pa (the_enat (epred (llength xsa)) + ka) (LCons (lhd ysa) LNil)"
       using a2 by simp
       then have f12: "lappend (lappend (kfilter Pa ka (lbutlast xsa)) 
                                        (kfilter Pa (ka + the_enat (llength (lbutlast xsa))) LNil)) 
                               (LCons (the_enat (epred (llength xsa)) + ka) LNil) = kfilter Pa ka xsa"
       using f11 f10 f9 a5 a4 a3 by (metis add.commute lappend_lbutlast_llast_id_lfinite llength_lbutlast)
       have "ltl (kfilter Pa (ka + the_enat (epred (llength xsa))) ysa) =
             kfilter Pa (Suc (the_enat (epred (llength xsa)) + ka)) (ltl ysa)"
       using f8 f6 by (simp add: add.commute)
       then show "kfilter Pa ka (lappend xsa (ltl ysa)) = 
                  lappend (kfilter Pa ka xsa) (ltl (kfilter Pa (ka + the_enat (epred (llength xsa))) ysa))"
       using f12 f11 f10 f9 f8 f7 f6 a5 a4 a3 
       by (metis add.commute lappend_lbutlast_llast_id_lfinite lappend_snocL1_conv_LCons2 llength_lbutlast)
       qed 
   show "\<And>xs ys P k x xa xb.
       \<not> lnull xs \<Longrightarrow>
       \<not> lnull ys \<Longrightarrow>
       llast xs = lhd ys \<Longrightarrow>
       P (lhd ys) \<Longrightarrow>
       lfinite xs \<Longrightarrow>
       \<forall>b. ys \<noteq> LCons b LNil \<Longrightarrow>
       \<forall>b. kfilter P (k + the_enat (epred (llength xs))) ys \<noteq> LCons b LNil \<Longrightarrow>
       x \<in> lset ys \<Longrightarrow>
       P x \<Longrightarrow>
       xa \<in> lset xs \<Longrightarrow>
       P xa \<Longrightarrow>
       P xb \<Longrightarrow>
       xb \<in> lset (ltl ys) \<Longrightarrow>
       kfilter P k (lappend xs (ltl ys)) = 
       lappend (kfilter P k xs) (ltl (kfilter P (k + the_enat (epred (llength xs))) ys)) "
       by (simp add: f2)
    qed   
qed

lemma nleast_conv:
 assumes "\<exists>x \<in> nset xs. P x"
 shows   " nleast P xs = (LEAST na. na \<le> nlength xs \<and> P (nnth xs na))"
using assms  
by transfer
   (auto simp add: min_def lleast_def,
    metis co.enat.exhaust_sel iless_Suc_eq llength_eq_0)

lemma nfilter_chop: 
assumes "nlast xs = nfirst ys"
        " P (nlast xs)"
        "nfinite xs" 
shows "nfilter P (nfuse xs ys) = nfuse (nfilter P xs)  (nfilter P ys)"
proof (cases "is_NNil ys")
case True
then show ?thesis by (metis nfilter_NNil nfuse_def nfuse_leftneutral) 
next
case False
then show ?thesis 
  proof (cases "is_NNil (nfilter P ys)")
  case True
  then show ?thesis unfolding nfuse_def using assms nfilter_nappend2[of "ntl ys" P xs] 
   nfilter_expand[of ys P]  
   by (auto simp add: nfirst_def nellist.case_eq_if )
    (metis nellist.discI(2) nellist.set_sel(2)  nset_nlast)
  next
  case False
  then show ?thesis 
    proof -
     have 1: "\<exists>x\<in>nset xs. P x "
       using assms nset_nlast by blast 
     have 2: "\<exists>x\<in>nset (ntl ys). P x "
       using False assms 
       by (metis nellist.collapse(1) nellist.collapse(2) nellist.disc(1) nfilter_NCons_a
         nfilter_NNil nnth_0 ntaken_0 ntaken_nlast)
     have 3: "\<not>is_NNil (nfilter P ys) \<Longrightarrow> 
               nfilter P (nappend xs (ntl ys)) = nappend (nfilter P xs) (nfilter P (ntl ys))"
       by (simp add: "1" "2" assms(3) nappend_nfilter_nfinite)
     have 4: "is_NNil (nfilter P ys) \<Longrightarrow> nfilter P xs = nappend (nfilter P xs) (nfilter P (ntl ys))"
       by (simp add: False)
     have 5: "nfilter P (nfuse xs ys) = nfilter P  (nappend xs (ntl ys))" 
       unfolding nfuse_def 
       by (metis (full_types) False nellist.collapse(1) nfilter_NNil)
     have 6: " (ntl (nfilter P ys)) =  (nfilter P (ntl ys))"
       using assms 2 False nfilter_expand[of "ys" P]
       by (metis  in_nset_ntlD nellist.case_eq_if nellist.sel(5) nfirst_def) 
     show ?thesis 
       by (metis "3" "5" "6" False nfuse_def)
   qed  
  qed    
qed


lemma nfilter_chop1:
assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
shows  "nfilter P xs = nfuse (nfilter P (ntaken n xs))  (nfilter P (ndropn n xs))"
using assms 
by (metis nfuse_ntaken_ndropn ndropn_nfirst nfilter_chop nfinite_ntaken ntaken_nlast)

lemma nfilter_chop1_ntaken:
assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
shows  "ntaken (the_enat (nlength(nfilter P (ntaken n xs)))) (nfilter P xs) = 
         (nfilter P (ntaken n xs))  "
using assms nfilter_nappend_ntaken by (metis nfinite_ntaken nset_nlast)

lemma nfilter_nlast:
 assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
 shows  "nlast(nfilter P (ntaken n xs)) = (nlast(ntaken n xs))"
using assms
proof (induction n arbitrary: xs)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
  proof (cases xs)
  case (NNil x1)
  then show ?thesis by auto
  next
  case (NCons x21 x22)
  then show ?thesis using Suc 
  by auto 
     (metis Suc_ile_eq iless_Suc_eq nfilter_NCons nfinite_ntaken nlast_NCons nlength_NCons 
      nset_nlast ntaken_Suc_NCons)
  qed
qed

lemma nfilter_nfirst:
 assumes " P (nfirst(ndropn n xs))"
 shows   "nfirst(nfilter P (ndropn n xs)) = (nfirst (ndropn n xs)) "
using assms
proof (induction n arbitrary: xs)
case 0
then show ?case 
  by (auto simp add: ndropn_nfirst)
     (metis nellist.set_intros(1) nfilter_NNil nfilter_nappend_ntaken nlast_NNil nlength_NNil
      ntaken_0 the_enat_0 zero_enat_def zero_le)
next
case (Suc n)
then show ?case 
   by (metis ndropn_ndropn plus_1_eq_Suc)
qed

lemma nfilter_chop1_ndropn:
assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
shows  "ndropn (the_enat (nlength(nfilter P (ntaken n xs)))) (nfilter P xs) = 
        (nfilter P (ndropn n xs))  "
proof -
 have 1: "(nfilter P xs) = nfuse (nfilter P (ntaken n xs))  (nfilter P (ndropn n xs))"
   by (simp add: assms nfilter_chop1) 
 have 2: "nlast(nfilter P (ntaken n xs)) = nfirst (nfilter P (ndropn n xs))"
   by (metis assms(1) assms(2) ndropn_nfirst nfilter_nfirst nfilter_nlast ntaken_nlast)
 have 3: "ndropn (the_enat (nlength(nfilter P (ntaken n xs)))) 
                 (nfuse (nfilter P (ntaken n xs))  (nfilter P (ndropn n xs))) =
          (nfilter P (ndropn n xs))"
   by (metis "2" assms(1) enat_the_enat infinity_ileE length_nfilter_le min_absorb1 ndropn_nfuse 
       nfinite_conv_nlength_enat ntaken_nlength) 
 show ?thesis by (simp add: "1" "3")        
qed

lemma nkfilter_chop1:
  assumes "(enat n) \<le> nlength xs"
          " P (nlast (ntaken n xs))"
  shows "nkfilter P k xs = nfuse (nkfilter P k (ntaken n xs)) (nkfilter P (k+n) (ndropn n xs))"
using assms nfuse_ntaken_ndropn[of n xs] nkfilter_chop[of "(ntaken n xs)" "(ndropn n xs)" P k]
by (metis min.orderE ndropn_nfirst nfinite_ntaken ntaken_nlast ntaken_nlength the_enat.simps) 

lemma nkfilter_nlast:
 assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
 shows  "nlast(nkfilter P k (ntaken n xs)) = k+n"
using assms
proof (induction n arbitrary: k xs)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
 proof (cases xs)
 case (NNil x1)
 then show ?thesis
 using Suc.prems(1) enat_0_iff(1) by auto 
 next
 case (NCons x21 x22)
 then show ?thesis 
   using Suc 
   proof auto
   assume a1: "P (nlast (ntaken n x22))"
   assume a2: "\<And>xs k. \<lbrakk>enat n \<le> nlength xs; P (nlast (ntaken n xs))\<rbrakk> \<Longrightarrow> 
               nlast (nkfilter P k (ntaken n xs)) = k + n"
   assume "enat (Suc n) \<le> eSuc (nlength x22)"
   then have "enat n < eSuc (nlength x22)"
   using Suc_ile_eq by blast
   then have f3: "enat n \<le> nlength x22"
   by (meson iless_Suc_eq)
   have "\<exists>a. a \<in> nset (ntaken n x22) \<and> P a"
   using a1 nfinite_ntaken nset_nlast by blast
   then show "nlast (nkfilter P k (NCons x21 (ntaken n x22))) = Suc (k + n)"
   using f3 a2 a1 by (simp add: Bex_def_raw)
   qed 
 qed
qed

lemma nkfilter_nfirst:
 assumes " P (nfirst(ndropn n xs))"
 shows   "nfirst(nkfilter P k (ndropn n xs)) = k "
 using assms
proof (induction n arbitrary: k xs)
case 0
then show ?case 
 by (metis enat_defs(1) nellist.set_intros(1) nkfilter_NNil nkfilter_nappend_ntaken nlength_NNil 
     nnth_NNil ntaken_0 the_enat.simps zero_le)
next
case (Suc n)
then show ?case 
proof (cases xs)
case (NNil x1)
then show ?thesis using Suc
by (metis ndropn_ndropn plus_1_eq_Suc) 
next
case (NCons x21 x22)
then show ?thesis using Suc
by auto 
qed
qed
 
lemma nkfilter_chop1_ndropn:
assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
shows  "ndropn (the_enat (nlength(nkfilter P k (ntaken n xs)))) (nkfilter P k xs) = 
        (nkfilter P (k+n) (ndropn n xs))  "
proof -
 have 1: "(nkfilter P k xs) = nfuse (nkfilter P k (ntaken n xs))  (nkfilter P (k+n) (ndropn n xs))"
   by (simp add: assms nkfilter_chop1) 
 have 2: "nlast(nkfilter P k (ntaken n xs)) = nfirst (nkfilter P (k+n) (ndropn n xs))"
   by (metis assms(1) assms(2) ndropn_nfirst nkfilter_nfirst nkfilter_nlast ntaken_nlast)    
 have 3: "ndropn (the_enat (nlength(nkfilter P k (ntaken n xs)))) 
                 (nfuse (nkfilter P k (ntaken n xs))  (nkfilter P (k+n) (ndropn n xs))) =
          (nkfilter P (k+n) (ndropn n xs))"
   by (metis "2" assms(2) enat_the_enat infinity_ileE ndropn_nfuse nfinite_conv_nlength_enat 
       nfinite_ntaken nlength_nkfilter_le nset_nlast)
  show ?thesis by (simp add: "1" "3")        
qed

lemma nkfilter_chop1_ntaken:
assumes " n\<le> nlength xs"
        "  P (nlast (ntaken n xs))"
shows  "ntaken (the_enat (nlength(nkfilter P k (ntaken n xs)))) (nkfilter P k xs) = 
        (nkfilter P k (ntaken n xs))  "
proof -
 have 1: "(nkfilter P k xs) = nfuse (nkfilter P k (ntaken n xs))  (nkfilter P (k+n) (ndropn n xs))"
   by (simp add: assms nkfilter_chop1) 
 have 2: "nlast(nkfilter P k (ntaken n xs)) = nfirst (nkfilter P (k+n) (ndropn n xs))"
   by (metis assms(1) assms(2) ndropn_nfirst nkfilter_nfirst nkfilter_nlast ntaken_nlast)
 have 3: "ntaken (the_enat (nlength(nkfilter P k (ntaken n xs)))) 
                 (nfuse (nkfilter P k (ntaken n xs))  (nkfilter P (k+n) (ndropn n xs))) =
          (nkfilter P k (ntaken n xs))"
   by (metis "1" assms(1) assms(2) nfinite_ntaken nkfilter_nappend_ntaken nset_nlast)   
 show ?thesis by (simp add: "1" "3")        
qed


end

end